################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

from ast import Num
from audioop import mul
from pipes import Template
import struct
from pycutlass.library import DataTypeTag
from pycutlass import *
import cutlass
from scipy.special import erf

from pycutlass.c_types import MatrixCoord_
from pycutlass.frontend import NumpyFrontend

from cuda import cuda
from cuda import cudart

dtype2ctype = {
    cutlass.float16: ctypes.c_uint16,
    cutlass.float32: ctypes.c_float,
    cutlass.float64: ctypes.c_double,
    cutlass.int32: ctypes.c_int32
}


#################################################################################################
#
# Epilogue Functors
#
#################################################################################################

class EpilogueFunctorBase:
    """
    Base class for thread-level epilogue functors
    """
    def __init__(self) -> None:
        pass
    
    def emit(self, tag, template_argument):
        template = """${tag}<${arguments}>"""
        arguments = ""
        for idx, arg in enumerate(template_argument):
            arguments += arg
            if idx < len(template_argument) - 1:
                arguments += ", "
        values = {
            "tag": tag,
            "arguments": arguments
        }

        return SubstituteTemplate(template, values)
        


class LinearCombination(EpilogueFunctorBase):
    """
    Apply a linear combination operator to an array of elements
    D = alpha * accumulator + beta * source

    :param element_output: data type used to load and store tensors
    
    :param epilogue_vector_length: number of elements computed per operation. 
    Usually it is 128/sizeof_bits<ElementOutput_>, but we use 64 and 32 sometimes
    when there are not enough data to store

    :param element_accumulator: Accumulator data type

    :param element_epilogue: data type used to compute linear combination
    """
    tag = "cutlass::epilogue::thread::LinearCombination"
    def __init__(
        self, element_output, epilogue_vector_length, 
        element_accumulator=None, element_epilogue=None) -> None: # TODO bind ScaleType
        super().__init__()

        if element_accumulator is None:
            element_accumulator = element_output
        if element_epilogue is None:
            element_epilogue = element_output
        
        self.element_output = element_output
        self.element_accumulator = element_accumulator
        self.element_epilogue = element_epilogue

        self.template_arguments = [
            DataTypeTag[element_output], str(epilogue_vector_length),
            DataTypeTag[element_accumulator], DataTypeTag[element_epilogue]
        ]

        # get epilogue output op type
        c_element_epilogue = dtype2ctype[self.element_epilogue]
        element_epilogue = self.element_epilogue

        class _EpilogueOutputOpParams(ctypes.Structure):
            _fields_ = [
                ("alpha_data", ctypes.c_longlong*2),
                ("beta_data", ctypes.c_longlong*2),
                ("alpha", c_element_epilogue),
                ("beta", c_element_epilogue),
                ("alpha_ptr", ctypes.c_void_p),
                ("beta_ptr", ctypes.c_void_p),
            ]
            def __init__(self, alpha, beta, *args) -> None:
                self.alpha = element_epilogue(alpha).storage
                self.beta = element_epilogue(beta).storage
        self.epilogue_type = _EpilogueOutputOpParams
    
    def emit(self):
        return super().emit(self.tag, self.template_arguments)


class LinearCombinationClamp(LinearCombination):
    """
    Applies a linear combination operator to an array of elements then clamps 
    the output before converting to the output element type.

    D = alpha * accumulator + beta * source + uniform

    :param element_output: data type used to load and store tensors
    
    :param epilogue_vector_length: number of elements computed per operation. 
    Usually it is 128/sizeof_bits<ElementOutput_>, but we use 64 and 32 sometimes
    when there are not enough data to store

    :param element_accumulator: Accumulator data type

    :param element_epilogue: data type used to compute linear combination
    """
    tag = "cutlass::epilogue::thread::LinearCombinationClamp"
    def __init__(
        self, element_output, epilogue_vector_length, 
        element_accumulator=None, element_epilogue=None) -> None:
        # Base constructor
        super().__init__(
            element_output, epilogue_vector_length, 
            element_accumulator, element_epilogue)
        
        c_element_epilogue = dtype2ctype[self.element_epilogue]
        element_epilogue = self.element_epilogue

        class _EpilogueOutputOpParams(ctypes.Structure):
            _fields_ = [
                ("alpha", c_element_epilogue),
                ("beta", c_element_epilogue),
                ("alpha_ptr", ctypes.c_void_p),
                ("beta_ptr", ctypes.c_void_p),
            ]
            def __init__(self, alpha, beta, *args) -> None:
                self.alpha = element_epilogue(alpha).storage
                self.beta = element_epilogue(beta).storage
        self.epilogue_type =  _EpilogueOutputOpParams


class FastLinearCombinationClamp(EpilogueFunctorBase):
    """
    Applies a linear combination operator to an array of elements then clamps
    the output before converting to the output element type.

    D = alpha * accumulator + beta * source

    Note: The below method only when problem_size_K <= 256 for signed int8 gemm
    or problem_size_K <= 128 for unsigned int8 gemm. The default approach is
    above.

    :param element_output: data type used to load and store tensors
    
    :param epilogue_vector_length: number of elements computed per operation. 
    Usually it is 128/sizeof_bits<ElementOutput_>, but we use 64 and 32 sometimes
    when there are not enough data to store
    """
    tag = "cutlass::epilogue::thread::FastLinearCombinationClamp"
    def __init__(self, element_output, epilogue_vector_length, *args) -> None:
        super().__init__()

        self.template_arguments = [
            DataTypeTag[element_output], str(epilogue_vector_length)
        ]

        self.element_accumulator = cutlass.int32
        self.element_epilogue = cutlass.float32

        # get epilogue output op
        c_element_epilogue = dtype2ctype[self.element_epilogue]
        element_epilogue = self.element_epilogue

        class _EpilogueOutputOpParams(ctypes.Structure):
            _fields_ = [
                ("alpha", c_element_epilogue),
                ("beta", c_element_epilogue),
                ("alpha_ptr", ctypes.c_void_p),
                ("beta_ptr", ctypes.c_void_p),
            ]
            def __init__(self, alpha, beta, *args) -> None:
                self.alpha = element_epilogue(alpha).storage
                self.beta = element_epilogue(beta).storage
        self.epilogue_type =  _EpilogueOutputOpParams
    
    def emit(self):
        return super().emit(self.tag, self.template_arguments)


class LinearCombinationGeneric(LinearCombination):
    """
    Applies a linear combination operator followed by an activation function 
    to an array of elements.

    D = activation(alpha * accumulator + beta * source)

    :param activation_functor: input activation functor

    :param element_output: data type used to load and store tensors
    
    :param epilogue_vector_length: number of elements computed per operation. 
    Usually it is 128/sizeof_bits<ElementOutput_>, but we use 64 and 32 sometimes
    when there are not enough data to store

    :param element_accumulator: Accumulator data type

    :param element_epilogue: data type used to compute linear combination
    """
    tag = "cutlass::epilogue::thread::LinearCombinationGeneric"
    def __init__(
        self, activation_functor,
        element_output, epilogue_vector_length, 
        element_accumulator=None, element_epilogue=None) -> None:
        super().__init__(
            element_output, epilogue_vector_length, 
            element_accumulator, element_epilogue)
        
        self.template_arguments = [
            activation_functor.emit(),] + self.template_arguments
        
        self.activation_functor = activation_functor
        self.element_epilogue = element_epilogue
    
        # get epilogue output op
        self.epilogue_type = self.activation_functor.epilogue_output_op(self.element_epilogue)


class ActivationFunctor:
    """
    Base class for frequently used activation functions
    """
    def __init__(self, element_compute) -> None:
        pass
    @staticmethod
    def numpy(x: np.ndarray):
        raise NotImplementedError()

    def emit(self):
        return self.tag
    
    @staticmethod
    def epilogue_output_op(element_epilogue):
        c_element_epilogue = dtype2ctype[element_epilogue]

        class _EpilogueOutputOpParams(ctypes.Structure):
            _fields_ = [
                ("alpha", c_element_epilogue),
                ("beta", c_element_epilogue),
                ("alpha_ptr", ctypes.c_void_p),
                ("beta_ptr", ctypes.c_void_p),
            ]
            def __init__(self, alpha, beta, *args) -> None:
                self.alpha = element_epilogue(alpha).storage
                self.beta = element_epilogue(beta).storage
        return _EpilogueOutputOpParams

# identity operator
class identity(ActivationFunctor):
    def numpy(x: np.ndarray):
        return x

# ReLu operator, 
class relu(ActivationFunctor):
    tag = "cutlass::epilogue::thread::ReLu"

    def __init__(self, element_compute):
        super().__init__(element_compute)
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("threshold", dtype2ctype[element_compute])
            ]
            def __init__(self, threshold=0.) -> None:
                self.threshold = element_compute(threshold).storage
        self.argument_type = _Arguments
    
    def emit_visitor(self):
        return "cutlass::ReLUVisitor"
    
    @staticmethod
    def numpy(x: np.ndarray):
        return np.maximum(x, 0)

# Leaky ReLu operator
class leaky_relu(ActivationFunctor):
    tag = "cutlass::epilogue::thread::LeakyReLU"

    def __init__(self, element_compute) -> None:
        super().__init__(element_compute)
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("leaky_alpha", dtype2ctype[element_compute])
            ]
            def __init__(self, leaky_alpha) -> None:
                self.leaky_alpha = element_compute(leaky_alpha).storage
        self.argument_type = _Arguments
    
    def emit_visitor(self):
        return "cutlass::LeakyReLUVisitor"

    @staticmethod
    def numpy(x: np.ndarray, leaky_alpha):
        return np.maximum(x, 0) + np.minimum(x, 0) * leaky_alpha
    
    def epilogue_output_op(self, element_epilogue):
        c_element_epilogue = dtype2ctype[element_epilogue]
        class _EpilogueOutputOpParams(ctypes.Structure):
            _fields_ = [
                ("alpha", c_element_epilogue),
                ("beta", c_element_epilogue),
                ("alpha_ptr", ctypes.c_void_p),
                ("beta_ptr", ctypes.c_void_p),
                ("leaky_alpha", c_element_epilogue)
            ]
            def __init__(self, alpha, beta, leaky_alpha=0.2, *args) -> None:
                self.alpha = element_epilogue(alpha).storage
                self.beta = element_epilogue(beta).storage
                self.alpha_ptr = 0
                self.beta_ptr = 0
                self.leaky_alpha = element_epilogue(leaky_alpha).storage
        return _EpilogueOutputOpParams

# Tanh operator
class tanh(ActivationFunctor):
    tag = "cutlass::epilogue::thread::Tanh"

    def __init__(self, element_compute) -> None:
        super().__init__(element_compute)
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("tmp", ctypes.c_int)
            ]
            def __init__(self, *args) -> None:
                self.tmp = 0
        self.argument_type = _Arguments
    
    def emit_visitor(self):
        return "cutlass::TanhVisitor"

    @staticmethod
    def numpy(x: np.ndarray):
        return np.tanh(x)

def sigmoid_op(x: np.ndarray):
    return 1. / (1. + np.exp(-x))

# Sigmoid operator
class sigmoid(ActivationFunctor):
    tag = "cutlass::epilogue::thread::Sigmoid"

    @staticmethod
    def numpy(x: np.ndarray):
        return sigmoid_op(x)

# SiLu operator
class silu(ActivationFunctor):
    tag = "cutlass::epilogue::thread::SiLu"

    @staticmethod
    def numpy(x: np.ndarray):
        return x * sigmoid_op(x)

# Hardswish operator
class hardswish(ActivationFunctor):
    tag = "cutlass::epilogue::thread::HardSwish"

    @staticmethod
    def numpy(x: np.ndarray):
        relu6 = np.minimum(np.maximum(x + 3., 0), 6.)
        return x * relu6 / 6.

# GELU operator
class gelu(ActivationFunctor):
    tag = "cutlass::epilogue::thread::GELU"

    @staticmethod
    def numpy(x: np.ndarray):
        return 0.5 * x * (1 + erf(x / np.sqrt(2.)))

# reduction operator
def reduction_op(tensor, direction, math, factor):
    batch, m, n = tensor.shape
    if math == "Add":
        if direction == "row":
            num_cta_n = (n + factor - 1) // factor
            reduction = np.transpose(
                np.sum(tensor.reshape(batch, m, num_cta_n, factor), axis=-1), 
                axes=[0, 2, 1]).flatten()
        elif direction == "column":
            num_cta_m = (m + factor - 1) // factor
            reduction = np.sum(
                tensor.reshape(batch, num_cta_m, factor, n), axis=-2).flatten()
        else:
            raise NotImplementedError
        return  reduction
    else:
        raise NotImplementedError

# # GELU operator implemented using the taylor series approximation
# class GELU_taylor(ActivationFunctor):
#     tag = "cutlass::epilogue::thread::GELU_taylor"

# # Computes backwards pass for GELU operator
# class dGELU(ActivationFunctor):
#     tag = "cutlass::epilogue::thread::dGELU"

################################################################################
# Epilogue Visitor
################################################################################


class LayerNorm(EpilogueFunctorBase):
    """
    Apply a linear combination operator to an array of elements
    D = alpha * accumulator + beta * source

    :param element_output: data type used to load and store tensors
    
    :param epilogue_vector_length: number of elements computed per operation. 
    Usually it is 128/sizeof_bits<ElementOutput_>, but we use 64 and 32 sometimes
    when there are not enough data to store

    :param element_accumulator: Accumulator data type

    :param element_epilogue: data type used to compute linear combination
    """
    KernelTemplate = """

cutlass::epilogue::threadblock::EpilogueVisitorLayerNorm<
    cutlass::gemm::GemmShape<${threadblock_shape_m}, ${threadblock_shape_n}, ${threadblock_shape_k}>,
    ${operation_name}_default::kThreadCount,
    ${operation_name}_default::Epilogue::OutputTileIterator,
    ${operation_name}_default::Epilogue::AccumulatorFragmentIterator::AccumulatorTile,
    ${element_compute}, // element_compute
    ${element_variance}, // element_variance
    ${element_mean}, // element_mean
    ${element_layer_norm_compute}, // element_layer_norm_compute
    ${epilogue_functor},
    ${shifted_k}>;
"""
    headers = ["gemm/gemm_universal_with_visitor.h",
                "epilogue/epilogue_visitor_with_layernorm.h"]
    def __init__(
        self, elementwise_functor,
        element_variance=None, element_mean=None, 
        element_layer_norm_compute=None, shifted_k=True) -> None: # TODO bind ScaleType
        super().__init__()

        self.elementwise_functor = elementwise_functor
        self.element_compute = elementwise_functor.element_epilogue
        self.element_output = elementwise_functor.element_output
        
        if element_variance is None:
            self.element_variance = self.element_output
        if element_mean is None:
            self.element_mean = self.element_output
        if element_layer_norm_compute is None:
            self.element_layer_norm_compute = self.element_compute
        if shifted_k:
            self.shifted_k = "true"
        else:
            self.shifted_k = "false"
        
        # get epilogue output op
        elementwise_params_type = self.elementwise_functor.epilogue_type
        
        class _EpilogueVisitorParams(ctypes.Structure):
            _fields_ = [
                ("element_wise", elementwise_params_type),
                ("ptr_Variance", ctypes.c_void_p),
                ("ptr_Mean_", ctypes.c_void_p),
                ("ptr_Shifted_K_", ctypes.c_void_p),
                ("extent", MatrixCoord_)
            ]
            def __init__(self, elementwise_params, variance, mean, shift_k, extent) -> None:
                self.element_wise = elementwise_params
                if isinstance(variance, np.ndarray):
                    self.buffer_variance = NumpyFrontend.argument(variance, False)
                    self.buffer_mean = NumpyFrontend.argument(mean, False)
                    self.buffer_shift_k = NumpyFrontend.argument(shift_k, False)
                    self.ptr_Variance = int(self.buffer_variance.ptr)
                    self.ptr_Mean_ = int(self.buffer_mean.ptr)
                    self.ptr_Shifted_K_ = int(self.buffer_shift_k.ptr)
                    self.extent = MatrixCoord_(extent[0], extent[1])

                    self.host_variance = variance
                    self.host_mean = mean
                    self.host_shift_k = shift_k
            
            def sync(self, stream_sync=True):
                if stream_sync:
                    err, = cudart.cudaDeviceSynchronize()
                    if err != cuda.CUresult.CUDA_SUCCESS:
                        raise RuntimeError("CUDA Error %s" % str(err))
                
                # if hasattr(self, "host_variance"):
                err, = cuda.cuMemcpyDtoH(
                    self.host_variance, cuda.CUdeviceptr(self.ptr_Variance), 
                    self.host_variance.size * self.host_variance.itemsize)
                err, = cuda.cuMemcpyDtoH(
                    self.host_mean, cuda.CUdeviceptr(self.ptr_Mean_), 
                    self.host_mean.size * self.host_mean.itemsize)
                err, = cuda.cuMemcpyDtoH(
                    self.host_shift_k, cuda.CUdeviceptr(self.ptr_Shifted_K_), 
                    self.host_shift_k.size * self.host_shift_k.itemsize)
                if err != cuda.CUresult.CUDA_SUCCESS:
                    raise RuntimeError("CUDA Error %s" % str(err))

        self.epilogue_type =  _EpilogueVisitorParams

    def emit(self, operation):
        values = {
            'threadblock_shape_m': str(operation.tile_description.threadblock_shape[0]),
            'threadblock_shape_n': str(operation.tile_description.threadblock_shape[1]),
            'threadblock_shape_k': str(operation.tile_description.threadblock_shape[2]),
            'operation_name': operation.procedural_name(),
            'element_compute': DataTypeTag[self.element_compute],
            'element_variance': DataTypeTag[self.element_variance],
            'element_mean': DataTypeTag[self.element_mean],
            'element_layer_norm_compute': DataTypeTag[self.element_layer_norm_compute],
            'epilogue_functor': self.elementwise_functor.emit(),
            'shifted_k': self.shifted_k
        }
        return SubstituteTemplate(self.KernelTemplate, values)



class AccumulatorOp:
    Template = """
using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpAccumulator<${element_accumulator}, ${elements_per_access}>;
"""
    counter = 0
    def __init__(self, element_accumulator, elements_per_access) -> None:
        self.element_accumulator = element_accumulator
        self.elements_per_access = elements_per_access

        self.instance_name = "AccumulatorOp%d" % AccumulatorOp.counter
        AccumulatorOp.counter += 1


        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("tmp", ctypes.c_int)
            ]
            def __init__(self):
                self.tmp = 0
        
        self.argument_type = _Arguments
    
    def emit(self, *args):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "elements_per_access": str(self.elements_per_access)
        }
        return SubstituteTemplate(self.Template, values)


class LinearCombinationOp:
    Template = """
${visitor_a}

${visitor_b}

using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpLinearCombination<
    ${element_accumulator}, ${element_compute}, 
    ${elements_per_access}, ${visitor_a_name}, ${visitor_b_name}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_compute, 
        elements_per_access, visitor_a, visitor_b) -> None:
        #
        self.element_accumulator = element_accumulator
        self.element_compute = element_compute
        self.elements_per_access = elements_per_access
        self.visitor_a = visitor_a
        self.visitor_b = visitor_b
        
        self.instance_name = "LinearCombinationOp%d" % LinearCombinationOp.counter
        LinearCombinationOp.counter += 1

        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("alpha", dtype2ctype[self.element_compute]),
                ("beta", dtype2ctype[self.element_compute]),
                ("visitor_a", self.visitor_a.argument_type),
                ("visitor_b", self.visitor_b.argument_type)
            ]
            def __init__(self, alpha, beta, visitor_a_arg, visitor_b_arg) -> None:
                self.alpha = element_compute(alpha).storage
                self.beta = element_compute(beta).storage
                self.visitor_a = visitor_a_arg
                self.visitor_b = visitor_b_arg
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_compute": DataTypeTag[self.element_compute],
            "elements_per_access": str(self.elements_per_access),
            "visitor_a_name": self.visitor_a.instance_name,
            "visitor_b_name": self.visitor_b.instance_name,
            "visitor_a": self.visitor_a.emit(operation),
            "visitor_b": self.visitor_b.emit(operation)
        }
        return SubstituteTemplate(self.Template, values)

class VectorAdd:
    def __init__(self, *args) -> None:
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("tmp", ctypes.c_int)
            ]
            def __init__(self, *args) -> None:
                self.tmp = 0
        self.argument_type = _Arguments

    def emit(self):
        return "cutlass::VectorAdd"

class VectorMult:
    def __init__(self, *args) -> None:
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("tmp", ctypes.c_int)
            ]
            def __init__(self, *args) -> None:
                self.tmp = 0
        self.argument_type = _Arguments

    def emit(self):
        return "cutlass::VectorMult"
        

class BinaryOp:
    Template = """
${visitor_a}

${visitor_b}

using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpBinary<
    ${element_accumulator}, ${element_compute}, 
    ${elements_per_access}, ${visitor_a_name}, ${visitor_b_name}, ${binary_op}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_compute, 
        elements_per_access, visitor_a, visitor_b, binary_op) -> None:
        #
        self.element_accumulator = element_accumulator
        self.element_compute = element_compute
        self.elements_per_access = elements_per_access
        self.visitor_a = visitor_a
        self.visitor_b = visitor_b
        self.binary_op = binary_op

        self.instance_name = "BinaryOp%d" % BinaryOp.counter
        BinaryOp.counter += 1

        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("binary_param", binary_op.argument_type),
                ("visitor_a", self.visitor_a.argument_type),
                ("visitor_b", self.visitor_b.argument_type)
            ]
            def __init__(self, binary_param, visitor_a_arg, visitor_b_arg) -> None:
                self.binary_param = binary_param
                self.visitor_a = visitor_a_arg
                self.visitor_b = visitor_b_arg
        
        self.argument_type = _Arguments
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_compute": DataTypeTag[self.element_compute],
            "elements_per_access": str(self.elements_per_access),
            "visitor_a_name": self.visitor_a.instance_name,
            "visitor_b_name": self.visitor_b.instance_name,
            "visitor_a": self.visitor_a.emit(operation),
            "visitor_b": self.visitor_b.emit(operation),
            "binary_op": self.binary_op.emit()
        }
        return SubstituteTemplate(self.Template, values)


class Mult:
    def __init__(self, element_compute) -> None:
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("alpha", dtype2ctype[element_compute])
            ]
            def __init__(self, alpha) -> None:
                self.alpha = element_compute(alpha).storage
        
        self.argument_type = _Arguments
    
    def emit_visitor(self):
        return "cutlass::Mult"

class UnaryOp:
    Template = """
${visitor}

using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpUnary<
    ${element_accumulator}, ${element_compute},
    ${elements_per_access}, ${visitor_name}, ${unary_op}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_compute,
        elements_per_access, visitor, unary_op) -> None:
        #
        self.element_accumulator = element_accumulator
        self.element_compute = element_compute
        self.elements_per_access = elements_per_access
        self.visitor = visitor
        self.unary_op = unary_op

        self.instance_name = "UnaryOp%d" % UnaryOp.counter
        UnaryOp.counter += 1

        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("unary_param", unary_op.argument_type),
                ("visitor_arg", self.visitor.argument_type)
            ]
            def __init__(self, unary_param, visitor_arg) -> None:
                self.unary_param = unary_param
                self.visitor_arg = visitor_arg
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_compute": DataTypeTag[self.element_compute],
            "elements_per_access": str(self.elements_per_access),
            "visitor_name": self.visitor.instance_name,
            "unary_op": self.unary_op.emit_visitor(),
            "visitor": self.visitor.emit(operation)
        }
        return SubstituteTemplate(self.Template, values)



class RowBroadcastOp:
    Template = """
using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpRowBroadcast<
    ${element_accumulator}, ${element_fragment}, ${input_tile_iterator}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_fragment) -> None:
        self.element_accumulator = element_accumulator
        self.element_fragment = element_fragment

        self.instance_name = "RowBroadcastOp%d" % RowBroadcastOp.counter
        RowBroadcastOp.counter += 1
        
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("broadcast_ptr", ctypes.c_void_p),
                ("batch_stride", ctypes.c_longlong)
            ]
            def __init__(self, broadcast_ptr, batch_stride=0):
                self.broadcast_ptr = int(broadcast_ptr)
                self.batch_stride = batch_stride
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_fragment": DataTypeTag[self.element_fragment],
            "input_tile_iterator": operation.procedural_name() + "_default::Epilogue::OutputTileIterator"
        }
        return SubstituteTemplate(self.Template, values)


class ColumnBroadcastOp:
    Template = """
using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpColumnBroadcast<
    ${element_accumulator}, ${element_fragment}, ${input_tile_iterator}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_fragment) -> None:
        self.element_accumulator = element_accumulator
        self.element_fragment = element_fragment

        self.instance_name = "ColumnBroadcastOp%d" % ColumnBroadcastOp.counter
        ColumnBroadcastOp.counter += 1
        
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("broadcast_ptr", ctypes.c_void_p),
                ("batch_stride", ctypes.c_longlong)
            ]
            def __init__(self, broadcast_ptr, batch_stride=0):
                self.broadcast_ptr = int(broadcast_ptr)
                self.batch_stride = batch_stride
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_fragment": DataTypeTag[self.element_fragment],
            "input_tile_iterator": operation.procedural_name() + "_default::Epilogue::OutputTileIterator"
        }
        return SubstituteTemplate(self.Template, values)


class TensorInputOp:
    Template = """
using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpTensorInput<
    ${element_accumulator}, ${input_tile_iterator}>;
"""
    counter = 0
    def __init__(self, element_accumulator) -> None:
        self.element_accumulator = element_accumulator

        self.instance_name = "TensorInputOp%d" % TensorInputOp.counter
        TensorInputOp.counter += 1
        
        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("input_ptr", ctypes.c_void_p),
                ("ldt", ctypes.c_int),
                ("batch_stride", ctypes.c_longlong)
            ]
            def __init__(self, input_ptr, ldt, batch_stride=0) -> None:
                self.input_ptr = int(input_ptr)
                self.ldt = ldt
                self.batch_stride = batch_stride
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "input_tile_iterator": operation.procedural_name() + "_default::Epilogue::OutputTileIterator"
        }
        return SubstituteTemplate(self.Template, values)

class TensorOutputOp:
    Template = """
${visitor}

using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpTensorOutput<
    ${element_accumulator}, ${output_tile_iterator}, ${visitor_name}>;
"""
    counter = 0
    def __init__(self, element_accumulator, visitor) -> None:
        self.element_accumulator = element_accumulator
        self.visitor = visitor

        self.instance_name = "TensorOutputOp%d" % TensorOutputOp.counter
        TensorOutputOp.counter += 1

        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("output_ptr", ctypes.c_void_p),
                ("ldt", ctypes.c_int),
                ("batch_stride", ctypes.c_longlong),
                ("visitor_arg", self.visitor.argument_type)
            ]
            def __init__(self, output_ptr, ldt, visitor_arg, batch_stride=0) -> None:
                self.output_ptr = int(output_ptr)
                self.ldt = int(ldt)
                self.visitor_arg = visitor_arg
                self.batch_stride = batch_stride
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "output_tile_iterator": operation.procedural_name() + "_default::Epilogue::OutputTileIterator",
            "visitor_name": self.visitor.instance_name,
            "visitor": self.visitor.emit(operation)
        }
        return SubstituteTemplate(self.Template, values)


class ColumnReductionOp:
    Template = """
${visitor}

using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpColumnReduction<
    cutlass::gemm::GemmShape<${threadblock_shape_m}, ${threadblock_shape_n}, ${threadblock_shape_k}>,
    ${element_accumulator}, ${element_reduction}, ${element_reduction_accumulator},
    ${output_tile_iterator}, ${visitor_name}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_reduction, 
        element_reduction_accumulator, visitor) -> None:
        self.element_accumulator = element_accumulator
        self.element_reduction = element_reduction
        self.element_reduction_accumulator = element_reduction_accumulator
        self.visitor = visitor

        self.instance_name = "ColumnReductionOp%d" % ColumnReductionOp.counter
        ColumnReductionOp.counter += 1

        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("reduction_ptr", ctypes.c_void_p),
                ("batch_stride", ctypes.c_longlong),
                ("visitor_arg", self.visitor.argument_type)
            ]
            def __init__(self, reduction_ptr, visitor_arg, batch_stride=0) -> None:
                self.reduction_ptr = reduction_ptr
                self.batch_stride = batch_stride
                self.visitor_arg = visitor_arg
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            'threadblock_shape_m': str(operation.tile_description.threadblock_shape[0]),
            'threadblock_shape_n': str(operation.tile_description.threadblock_shape[1]),
            'threadblock_shape_k': str(operation.tile_description.threadblock_shape[2]),
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_reduction": DataTypeTag[self.element_reduction],
            "element_reduction_accumulator": DataTypeTag[self.element_reduction_accumulator],
            "output_tile_iterator": operation.procedural_name() + "_default::Epilogue::OutputTileIterator",
            "visitor_name": self.visitor.instance_name,
            "visitor": self.visitor.emit(operation)
        }
        return SubstituteTemplate(self.Template, values)


class RowReductionOp:
    Template = """
${visitor}

using ${instance_name} = cutlass::epilogue::threadblock::VisitorOpRowReduction<
    cutlass::gemm::GemmShape<${threadblock_shape_m}, ${threadblock_shape_n}, ${threadblock_shape_k}>,
    ${element_accumulator}, ${element_reduction}, ${element_reduction_accumulator},
    ${output_tile_iterator}, ${visitor_name}>;
"""
    counter = 0
    def __init__(self, element_accumulator, element_reduction, 
        element_reduction_accumulator, visitor) -> None:
        self.element_accumulator = element_accumulator
        self.element_reduction = element_reduction
        self.element_reduction_accumulator = element_reduction_accumulator
        self.visitor = visitor

        self.instance_name = "RowReductionOp%d" % RowReductionOp.counter
        RowReductionOp.counter += 1

        class _Arguments(ctypes.Structure):
            _fields_ = [
                ("reduction_ptr", ctypes.c_void_p),
                ("batch_stride", ctypes.c_longlong),
                ("visitor_arg", self.visitor.argument_type)
            ]
            def __init__(self, reduction_ptr, visitor_arg, batch_stride=0) -> None:
                self.reduction_ptr = reduction_ptr
                self.visitor_arg = visitor_arg
                self.batch_stride = batch_stride
        
        self.argument_type = _Arguments
    
    def emit(self, operation):
        values = {
            "instance_name": self.instance_name,
            'threadblock_shape_m': str(operation.tile_description.threadblock_shape[0]),
            'threadblock_shape_n': str(operation.tile_description.threadblock_shape[1]),
            'threadblock_shape_k': str(operation.tile_description.threadblock_shape[2]),
            "element_accumulator": DataTypeTag[self.element_accumulator],
            "element_reduction": DataTypeTag[self.element_reduction],
            "element_reduction_accumulator": DataTypeTag[self.element_reduction_accumulator],
            "output_tile_iterator": operation.procedural_name() + "_default::Epilogue::OutputTileIterator",
            "visitor_name": self.visitor.instance_name,
            "visitor": self.visitor.emit(operation)
        }
        return SubstituteTemplate(self.Template, values)



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

## Test case generator for SM80

import pycutlass
from pycutlass import *
from pycutlass.test import *
import unittest

#
# Create GEMM operation
#

def TestGemmOperator(gemm_kind, math_inst, layout, alignment, tiling, arch, mixed=False,
    epilogue_functor=None, swizzling_functor=cutlass.IdentitySwizzle1, **kwargs):
    """
    Test GEMM Operation based on configuration
    """

    if "data_type" in kwargs.keys():
        data_type = kwargs["data_type"]
    else:
        if mixed or math_inst.element_a == cutlass.bfloat16:
            data_type = [
                math_inst.element_a,
                math_inst.element_b,
                math_inst.element_accumulator,
                math_inst.element_accumulator
            ]
        else:
            data_type = [
                math_inst.element_a,
                math_inst.element_b,
                math_inst.element_a,
                math_inst.element_accumulator
            ]
    
    tile_description = TileDescription(
        tiling[0], tiling[1], tiling[2],
        math_inst
    )

    A = TensorDescription(
        data_type[0], layout[0], alignment[0]
    )

    B = TensorDescription(
        data_type[1], layout[1], alignment[1]
    )

    C = TensorDescription(
        data_type[2], layout[2], alignment[2]
    )

    element_epilogue = data_type[3]
    if epilogue_functor is None:
        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)

    if gemm_kind == GemmKind.Universal:
        operation = GemmOperationUniversal(
            arch=arch, tile_description=tile_description,
            A=A, B=B, C=C,
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )
        if A.layout in [cutlass.ColumnMajorInterleaved32, cutlass.RowMajorInterleaved32]:
            return test_all_gemm(operation, "interleaved")
        else:
            return test_all_gemm(operation, "universal")
        
    elif gemm_kind == GemmKind.Grouped:
        operation = GemmOperationGrouped(
            arch, tile_description, A, B, C,
            epilogue_functor, swizzling_functor,
            precompute_mode=kwargs["precompute_mode"]
        )
        testbed = TestbedGrouped(operation=operation)
        return testbed.run(24)
    else:
        raise NotImplementedError("the gemm kind is not implemented")


def TestConv2dOperator(math_inst, alignment, tiling, arch, 
    stride_supports=[StrideSupport.Strided, StrideSupport.Strided, StrideSupport.Strided],
    epilogue_functor=None, 
    swizzling_functor=cutlass.IdentitySwizzle1, interleaved=False, **kwargs):
    """
    Test Conv2d Operation based on configurations
    """

    mixeds = [False, True, False]
    conv_kinds = [cutlass.conv.Operator.fprop, cutlass.conv.Operator.dgrad, cutlass.conv.Operator.wgrad]

    results = []

    default_swizzling_functor = swizzling_functor

    if "layout" in kwargs.keys():
        layout = kwargs["layout"]
    else:
        layout = (cutlass.TensorNHWC, cutlass.TensorNHWC, cutlass.TensorNHWC)

    for mixed, conv_kind, stride_support in zip(mixeds, conv_kinds, stride_supports):

        if "data_type" in kwargs.keys():
            data_type = kwargs["data_type"]
        else:
            if mixed or math_inst.element_a == cutlass.bfloat16:
                data_type = [
                    math_inst.element_a,
                    math_inst.element_b,
                    math_inst.element_accumulator,
                    math_inst.element_accumulator
                ]
            else:
                data_type = [
                    math_inst.element_a,
                    math_inst.element_b,
                    math_inst.element_a,
                    math_inst.element_accumulator
                ]
        # skip Int8 Conv Backward
        if data_type[0] == cutlass.int8 and conv_kind in [cutlass.conv.Operator.dgrad, cutlass.conv.Operator.wgrad]:
            continue

        A = TensorDescription(
            element=data_type[0],
            layout=layout[0],
            alignment=alignment[0])
        B = TensorDescription(
            element=data_type[1],
            layout=layout[1], 
            alignment=alignment[1])
        C = TensorDescription(
            element=data_type[2],
            layout=layout[2], 
            alignment=alignment[2])
        
        tile_description = TileDescription(
            threadblock_shape=tiling[0], stages=tiling[1], 
            warp_count=tiling[2],
            math_instruction=math_inst
        )

        if conv_kind == cutlass.conv.Operator.dgrad and stride_support == StrideSupport.Strided:
            swizzling_functor = cutlass.StridedDgradIdentitySwizzle1
        else:
            swizzling_functor = default_swizzling_functor
        
        if epilogue_functor is None:
            epilogue_functor_ = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, data_type[3])

        operation = Conv2dOperation(
            conv_kind=conv_kind, iterator_algorithm=cutlass.conv.IteratorAlgorithm.optimized,
            arch=arch, tile_description=tile_description, A=A, B=B, C=C, 
            stride_support=stride_support,
            epilogue_functor=epilogue_functor_,
            swizzling_functor=swizzling_functor
        )
        
        results.append(test_all_conv2d(operation, interleaved=interleaved))
    
    return results



class Test_SM80(unittest.TestCase):
    def test_SM80_TensorOp_16816(self):
        math_instructions = [
            MathInstruction(
                [16, 8, 16], cutlass.float16, cutlass.float16, cutlass.float32,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add
            ),
            MathInstruction(
                [16, 8, 16], cutlass.float16, cutlass.float16, cutlass.float16,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add
            ),
            MathInstruction(
                [16, 8, 16], cutlass.bfloat16, cutlass.bfloat16, cutlass.float32,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add
            )
        ]

        layouts = [
            (cutlass.RowMajor, cutlass.RowMajor, cutlass.RowMajor),
            (cutlass.ColumnMajor, cutlass.RowMajor, cutlass.RowMajor),
            (cutlass.RowMajor, cutlass.ColumnMajor, cutlass.RowMajor)
        ]

        alignments = [
            (8, 8, 8), (4, 8, 8), (8, 4, 8)
        ]

        tilings = [
            ([256, 128, 32], 3, [4, 2, 1]),
            ([64, 256, 32], 4, [1, 4, 1]),
            ([128, 64, 64], 3, [2, 2, 1])
        ]

        for math_inst, layout, alignment, tiling in zip(math_instructions, layouts, alignments, tilings):
            self.assertTrue(TestGemmOperator(GemmKind.Universal, math_inst, layout, alignment, tiling, 80, False))
            self.assertTrue(TestGemmOperator(GemmKind.Grouped, math_inst, layout, alignment, tiling, 80, True, precompute_mode=SchedulerMode.Host))
            stride_supports = [StrideSupport.Strided, StrideSupport.Strided, StrideSupport.Strided]
            results = TestConv2dOperator(math_inst, alignment, tiling, 80, stride_supports=stride_supports)
            for res in results:
                self.assertTrue(res)

    def test_SM80_TensorOp_1688(self):
        # tf32 is not supported by most of python environment. Skip the test
        self.assertTrue(True)
    
    def test_SM80_TensorOp_1688_fast_math(self):
        math_instructions = [
            MathInstruction(
                [16, 8, 8], cutlass.tfloat32, cutlass.tfloat32, cutlass.float32,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add
            ),
            MathInstruction(
                [16, 8, 8], cutlass.float16, cutlass.float16, cutlass.float32,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add_fast_f16
            ),
            MathInstruction(
                [16, 8, 8], cutlass.bfloat16, cutlass.bfloat16, cutlass.float32,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add_fast_bf16
            ),
            MathInstruction(
                [16, 8, 8], cutlass.float32, cutlass.float32, cutlass.float32,
                cutlass.OpClass.TensorOp, MathOperation.multiply_add_fast_f32
            )
        ]

        layouts = [
            (cutlass.RowMajor, cutlass.RowMajor, cutlass.ColumnMajor),
            (cutlass.RowMajor, cutlass.ColumnMajor, cutlass.ColumnMajor),
            (cutlass.ColumnMajor, cutlass.RowMajor, cutlass.ColumnMajor),
            (cutlass.ColumnMajor, cutlass.ColumnMajor, cutlass.RowMajor)
        ]
        alignments = [
            (4, 4, 4), (4, 2, 4), (2, 4, 4), (2, 2, 4)
        ]
        tilings = [
            ([128, 256, 16], 3, [4, 2, 1]),
            ([64, 256, 16], 4, [1, 4, 1]),
            ([128, 64, 32], 3, [2, 2, 1]),
            ([256, 64, 32], 3, [4, 2, 1])
        ]
        data_type = [
            cutlass.float32, cutlass.float32, cutlass.float32, cutlass.float32
        ]
        for math_inst, layout, alignment, tiling in zip(math_instructions, layouts, alignments, tilings):
            self.assertTrue(
                TestGemmOperator(
                    GemmKind.Universal, math_inst, layout, 
                    alignment, tiling, 80, False, data_type=data_type))
            self.assertTrue(
                TestGemmOperator(
                    GemmKind.Grouped, math_inst, layout, alignment, tiling, 80, 
                    True, precompute_mode=SchedulerMode.Device, data_type=data_type))
            stride_supports = [StrideSupport.Unity, StrideSupport.Strided, StrideSupport.Unity]
            results = TestConv2dOperator(math_inst, alignment, tiling, 80, stride_supports=stride_supports, data_type=data_type)
            for res in results:
                self.assertTrue(res)

    def test_SM80_TensorOp_884(self):
        math_inst = MathInstruction(
            [8, 8, 4], cutlass.float64, cutlass.float64, cutlass.float64,
            cutlass.OpClass.TensorOp, MathOperation.multiply_add
        )
        layout = (cutlass.ColumnMajor, cutlass.ColumnMajor, cutlass.ColumnMajor)
        alignment = (1, 1, 1)

        tiling = ([64, 256, 16], 3, [2, 4, 1])
        data_type = [cutlass.float64, cutlass.float64, cutlass.float64, cutlass.float64]
        self.assertTrue(TestGemmOperator(GemmKind.Universal, math_inst, layout, alignment, tiling, 80, False, data_type=data_type))
        self.assertTrue(TestGemmOperator(GemmKind.Grouped, math_inst, layout, alignment, tiling, 80, True, precompute_mode=SchedulerMode.Device, data_type=data_type))
        stride_supports = [StrideSupport.Unity, StrideSupport.Strided, StrideSupport.Unity]
        results = TestConv2dOperator(math_inst, alignment, tiling, 80, stride_supports=stride_supports, data_type=data_type)
        for res in results:
            self.assertTrue(res)
    
    def test_SM80_TensorOp_16832_TN(self):
        math_inst = MathInstruction(
            [16, 8, 32], cutlass.int8, cutlass.int8, cutlass.int32,
            cutlass.OpClass.TensorOp, MathOperation.multiply_add_saturate
        )
        layout = (cutlass.RowMajor, cutlass.ColumnMajor, cutlass.ColumnMajor)
        alignment = (16, 16, 4)
        alignment_mixed = (16, 16, 16)
        tiling = ([128, 256, 64], 3, [2, 4, 1])

        data_type = [cutlass.int8, cutlass.int8, cutlass.int32, cutlass.int32]
        data_type_mixed = [cutlass.int8, cutlass.int8, cutlass.int8, cutlass.float32]

        self.assertTrue(TestGemmOperator(GemmKind.Universal, math_inst, layout, alignment, tiling, 80, False, data_type=data_type))
        self.assertTrue(TestGemmOperator(GemmKind.Grouped, math_inst, layout, alignment_mixed, tiling, 80, True, precompute_mode=SchedulerMode.Device, data_type=data_type_mixed))
        stride_supports = [StrideSupport.Strided, StrideSupport.Strided, StrideSupport.Strided]
        results = TestConv2dOperator(math_inst, alignment, tiling, 80, stride_supports=stride_supports, data_type=data_type)
        for res in results:
            self.assertTrue(res)
    
    def test_SM80_Simt_f32(self):
        math_inst = MathInstruction(
            [1, 1, 1], cutlass.float32, cutlass.float32, cutlass.float32,
            cutlass.OpClass.Simt, MathOperation.multiply_add
        )
        layout = (cutlass.RowMajor, cutlass.RowMajor, cutlass.RowMajor)
        alignment = (1, 1, 1)

        tiling = ([128, 256, 8], 4, [2, 4, 1])
        data_type = [cutlass.float32, cutlass.float32, cutlass.float32, cutlass.float32]
        self.assertTrue(TestGemmOperator(GemmKind.Universal, math_inst, layout, alignment, tiling, 80, False, data_type=data_type))
        self.assertTrue(TestGemmOperator(GemmKind.Grouped, math_inst, layout, alignment, tiling, 80, True, precompute_mode=SchedulerMode.Host, data_type=data_type))
        stride_supports = [StrideSupport.Strided, StrideSupport.Strided, StrideSupport.Strided]
        results = TestConv2dOperator(math_inst, alignment, tiling, 80, stride_supports=stride_supports, data_type=data_type)
        for res in results:
            self.assertTrue(res)

    def test_SM80_Simt_f64(self):
        math_inst = MathInstruction(
            [1, 1, 1], cutlass.float64, cutlass.float64, cutlass.float64,
            cutlass.OpClass.Simt, MathOperation.multiply_add
        )
        layout = (cutlass.RowMajor, cutlass.RowMajor, cutlass.ColumnMajor)
        alignment = (1, 1, 1)

        tiling = ([64, 128, 8], 5, [2, 2, 1])
        data_type = [cutlass.float64, cutlass.float64, cutlass.float64, cutlass.float64]
        self.assertTrue(TestGemmOperator(GemmKind.Universal, math_inst, layout, alignment, tiling, 80, False, data_type=data_type))
        self.assertTrue(TestGemmOperator(GemmKind.Grouped, math_inst, layout, alignment, tiling, 80, True, precompute_mode=SchedulerMode.Device, data_type=data_type))
        stride_supports = [StrideSupport.Strided, StrideSupport.Strided, StrideSupport.Strided]
        results = TestConv2dOperator(math_inst, alignment, tiling, 80, stride_supports=stride_supports, data_type=data_type)
        for res in results:
            self.assertTrue(res)

    def test_SM80_TensorOp_16832_Interleaved(self):
        math_inst = MathInstruction(
            [16, 8, 32], cutlass.int8, cutlass.int8, cutlass.int32,
            cutlass.OpClass.TensorOp, MathOperation.multiply_add_saturate
        )

        layout = (cutlass.ColumnMajorInterleaved32, cutlass.RowMajorInterleaved32, cutlass.ColumnMajorInterleaved32)
        alignment_mixed = (16, 16, 8)
        tiling = ([256, 64, 64], 4, [4, 1, 1])
        data_type_mixed = [cutlass.int8, cutlass.int8, cutlass.int8, cutlass.float32]

        epilogue_functor = FastLinearCombinationClamp(
            data_type_mixed[2], alignment_mixed[2]
        )

        self.assertTrue(TestGemmOperator(GemmKind.Universal, math_inst, layout, alignment_mixed, tiling, 80, False, data_type=data_type_mixed, epilogue_functor=epilogue_functor))
        stride_supports = [StrideSupport.Strided, StrideSupport.Strided, StrideSupport.Strided]
        layout = [cutlass.TensorNC32HW32, cutlass.TensorC32RSK32, cutlass.TensorNC32HW32]
        results = TestConv2dOperator(math_inst, alignment_mixed, tiling, 80, stride_supports=stride_supports, data_type=data_type_mixed, layout=layout, interleaved=True)
        for res in results:
            self.assertTrue(res)

    def SM80_SparseTensorOp_16832(self):
        pass
    def SM80_PlanarComplexTensorOp_16816(self):
        pass
    def SM80_SparseTensorOp_16816_fast_math(self):
        pass
    def SM80_TensorOp_1688_complex(self):
        pass
    def SM80_TensorOp_1688_fast_fp32_math_complex(self):
        pass
    def SM80_TensorOp_1688_rank_k(self):
        pass
    def SM80_TensorOp_1688_rank_k_complex(self):
        pass
    def SM80_TensorOp_1688_trmm(self):
        pass
    def SM80_TensorOp_1688_trmm_complex(self):
        pass
    def SM80_TensorOp_1688_symm(self):
        pass
    def SM80_TensorOp_1688_symm_complex(self):
        pass
    def SM80_TensorOp_884_complex(self):
        pass
    def SM80_TensorOp_884_complex_gaussian(self):
        pass
    def SM80_TensorOp_884_rank_k(self):
        pass
    def SM80_TensorOp_884_rank_k_complex(self):
        pass
    def SM80_TensorOp_884_rank_k_complex_gaussian(self):
        pass
    def SM80_TensorOp_884_trmm(self):
        pass
    def SM80_TensorOp_884_trmm_complex(self):
        pass
    def SM80_TensorOp_884_trmm_complex_gaussian(self):
        pass
    def SM80_TensorOp_884_symm(self):
        pass
    def SM80_TensorOp_884_symm_complex(self):
        pass
    def SM80_TensorOp_884_symm_complex_gaussian(self):
        pass
    def SM80_SparseTensorOp_16864_TN(self):
        pass
    def SM80_TensorOp_16864_TN(self):
        pass
    def SM80_SparseTensorOp_168128_TN(self):
        pass
    def SM80_TensorOp_16864_Interleaved(self):
        pass
    def SM80_TensorOp_168256(self):
        pass
    def SM80_Simt_complex(self):
        pass


if __name__ == '__main__':
    pycutlass.get_memory_pool(2**20, 2**34)
    pycutlass.compiler.nvcc()
    unittest.main()



import pycutlass
from pycutlass import *
from pycutlass.test import *
import unittest

from pycutlass.test.gemm_testbed import test_all_gemm

class GemmF64TensorOpSm80(unittest.TestCase):
    def test_SM80_Device_Gemm_f64n_f64t_f64t_tensor_op_f64_32x32x16_16x16x16(self):
        math_inst = MathInstruction(
            instruction_shape=[8, 8, 4],
            element_a=cutlass.float64, element_b=cutlass.float64,
            element_accumulator=cutlass.float64, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[32, 32, 16],
            stages=4, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        # alignment 1 restricted for double
        A = TensorDescription(
            element=cutlass.float64, layout=cutlass.ColumnMajor,
            alignment=1
        )
        B = TensorDescription(
            element=cutlass.float64, layout=cutlass.RowMajor,
            alignment=1
        )
        C = TensorDescription(
            element=cutlass.float64, layout=cutlass.RowMajor,
            alignment=1
        )

        element_epilogue = cutlass.float64

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))
    
    def test_SM80_Device_Gemm_f64t_f64n_f64t_tensor_op_f64_64x64x16_32x32x16(self):
        math_inst = MathInstruction(
            instruction_shape=[8, 8, 4],
            element_a=cutlass.float64, element_b=cutlass.float64,
            element_accumulator=cutlass.float64, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        tile_description = TileDescription(
            threadblock_shape=[64, 64, 16],
            stages=4, warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        # alignment 1 restricted for double
        A = TensorDescription(
            element=cutlass.float64, layout=cutlass.RowMajor,
            alignment=1
        )
        B = TensorDescription(
            element=cutlass.float64, layout=cutlass.ColumnMajor,
            alignment=1
        )
        C = TensorDescription(
            element=cutlass.float64, layout=cutlass.RowMajor,
            alignment=1
        )

        element_epilogue = cutlass.float64

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, element_epilogue)
        
        swizzling_functor = cutlass.IdentitySwizzle1

        operation = GemmOperationUniversal(
            arch=80, tile_description=tile_description,
            A=A, B=B, C=C, 
            epilogue_functor=epilogue_functor, swizzling_functor=swizzling_functor
        )

        self.assertTrue(test_all_gemm(operation, "universal"))

if __name__ == '__main__':
    pycutlass.get_memory_pool(2**24, 2**24)
    unittest.main()



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################

import ctypes
from pycutlass.library import *

# 12B


class GemmCoord_(ctypes.Structure):
    _fields_ = [
        ("m", ctypes.c_int),
        ("n", ctypes.c_int),
        ("k", ctypes.c_int)
    ]

    def __init__(self, gemm_coord) -> None:
        for field_name, _ in self._fields_:
            setattr(self, field_name, getattr(gemm_coord, field_name)())


class MatrixCoord_(ctypes.Structure):
    _fields_ = [
        ("row", ctypes.c_int),
        ("column", ctypes.c_int)
    ]


dtype2ctype = {
    cutlass.float16: ctypes.c_uint16,
    cutlass.float32: ctypes.c_float,
    cutlass.float64: ctypes.c_double,
    cutlass.int32: ctypes.c_int32
}


def get_gemm_arguments(epilogue_functor):

    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _GemmArguments(ctypes.Structure):
        _fields_ = [
            ("mode", ctypes.c_int),
            ("problem_size", GemmCoord_),
            ("batch_count", ctypes.c_int),
            ("epilogue", _EpilogueOutputOpParams),
            ("ptr_A", ctypes.c_void_p),
            ("ptr_B", ctypes.c_void_p),
            ("ptr_C", ctypes.c_void_p),
            ("ptr_D", ctypes.c_void_p),
            ("batch_stride_A", ctypes.c_longlong),
            ("batch_stride_B", ctypes.c_longlong),
            ("batch_stride_C", ctypes.c_longlong),
            ("batch_stride_D", ctypes.c_longlong),
            ("stride_a", ctypes.c_longlong),
            ("stride_b", ctypes.c_longlong),
            ("stride_c", ctypes.c_longlong),
            ("stride_d", ctypes.c_longlong),
            ("lda", ctypes.c_longlong),
            ("ldb", ctypes.c_longlong),
            ("ldc", ctypes.c_longlong),
            ("ldd", ctypes.c_longlong),
            ("ptr_gather_A_indices", ctypes.c_void_p),
            ("ptr_gether_B_indices", ctypes.c_void_p),
            ("ptr_scatter_D_indices", ctypes.c_void_p)
        ]

    return _GemmArguments, _EpilogueOutputOpParams


###########################################################################################
# GEMM Grouped
###########################################################################################

# include/cutlass/gemm/kernel/gemm_grouped.h

def get_gemm_grouped_arguments(epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _GEMMGroupedArguments(ctypes.Structure):
        _fields_ = [
            ("problem_sizes", ctypes.c_void_p),
            ("problem_count", ctypes.c_int),
            ("threadblock_count", ctypes.c_int),
            ("output_op", _EpilogueOutputOpParams),
            ("ptr_A", ctypes.c_void_p),
            ("ptr_B", ctypes.c_void_p),
            ("ptr_C", ctypes.c_void_p),
            ("ptr_D", ctypes.c_void_p),
            ("lda", ctypes.c_void_p),
            ("ldb", ctypes.c_void_p),
            ("ldc", ctypes.c_void_p),
            ("ldd", ctypes.c_void_p),
            ("host_problem_sizes", ctypes.c_void_p)
        ]

    return _GEMMGroupedArguments, _EpilogueOutputOpParams

############################################################################################
# Convolution2D
############################################################################################


# We use the arguments as the interface


# include/cutlass/conv/conv2d_problem_size.h
# 64B
class Conv2DProblemSize(ctypes.Structure):
    _fields_ = [
        ("N", ctypes.c_int),
        ("H", ctypes.c_int),
        ("W", ctypes.c_int),
        ("C", ctypes.c_int),
        ("P", ctypes.c_int),
        ("Q", ctypes.c_int),
        ("K", ctypes.c_int),
        ("R", ctypes.c_int),
        ("S", ctypes.c_int),
        ("pad_h", ctypes.c_int),
        ("pad_w", ctypes.c_int),
        ("stride_h", ctypes.c_int),
        ("stride_w", ctypes.c_int),
        ("dilation_h", ctypes.c_int),
        ("dilation_w", ctypes.c_int),
        ("mode", ctypes.c_int),  # kCrossCorrelation: 0, kConvolution: 1
        ("split_k_slices", ctypes.c_int),
        ("groups", ctypes.c_int)
    ]

    def __init__(self, problem_size) -> None:
        for field_name, _ in self._fields_:
            setattr(self, field_name, getattr(problem_size, field_name))


# include/cutlass/layout/tensor.h
# 12B
class Layout4D(ctypes.Structure):
    _fields_ = [
        ("stride", ctypes.c_int * 3)
    ]

    def __init__(self, tensor_ref):
        stride = tensor_ref.stride()
        setattr(self, "stride", (stride.at(0), stride.at(1), stride.at(2)))

# TODO: Tensor 5-D takes ("stride", ctypes.c_int * 4)


# include/cutlass/conv/threadblock/conv2d_dgrad_filter_tile_access_iterator_optimized.h
# TensorRef is basically cutlass::TensorRef<Element, Layout>;
# include/cutlass/tensor_ref.h
# 24B
class TensorRef_(ctypes.Structure):
    _fields_ = [
        ("ptr", ctypes.c_void_p),
        ("layout", Layout4D)
    ]

    def __init__(self, tensor_ref):
        setattr(self, "ptr", tensor_ref.data())
        setattr(self, "layout", Layout4D(tensor_ref.layout()))


class TensorRef2D_(ctypes.Structure):
    _fields_ = [
        ("ptr", ctypes.c_void_p),
        ("stride", ctypes.c_int)
    ]


# include/cutlass/conv/kernel/implicit_gemm_convolution.h
# split_k_mode: kNone: 0, kSerial: 1, kParallel: 2, kParallelSerial: 3, kInvalid: 4

def get_conv2d_arguments(epilogue_functor):
    _EpilogueOutputOpParams = epilogue_functor.epilogue_type

    class _Conv2dArguments(ctypes.Structure):
        _fields_ = [
            ("problem_size", Conv2DProblemSize),  # 0
            ("ref_A", TensorRef_),  # 72
            ("ref_B", TensorRef_),  # 96
            ("ref_C", TensorRef_),  # 120
            ("ref_D", TensorRef_),  # 144
            ("output_op", _EpilogueOutputOpParams),  # 168
            ("split_k_mode", ctypes.c_int)  # 192
        ]

    return _Conv2dArguments, _EpilogueOutputOpParams


############################################################################################
# Reduction
############################################################################################


def get_reduction_params(epilogue_functor):
    _EpilogueOutputParams = epilogue_functor.epilogue_type

    class _ReductionParams(ctypes.Structure):
        _fields_ = [
            ("problem_size", MatrixCoord_),
            ("partitions", ctypes.c_int),
            ("partition_stride", ctypes.c_longlong),
            ("workspace", TensorRef2D_),
            ("destination", TensorRef2D_),
            ("source", TensorRef2D_),
            ("output_op", _EpilogueOutputParams)
        ]
    return _ReductionParams, _EpilogueOutputParams



#
# \file generator.py
#
# \brief Generates the CUTLASS Library's instances
#
#

import enum
import os.path
import shutil

from library import *

###################################################################################################

#
class Conv3dOperation:
  #
  def __init__(self, conv_kind, iterator_algorithm, arch, tile_description, A, B, C, element_epilogue, \
    stride_support, epilogue_functor = EpilogueFunctor.LinearCombination, swizzling_functor = SwizzlingFunctor.Identity4):

    self.operation_kind = OperationKind.Conv3d
    self.arch = arch
    self.tile_description = tile_description
    self.conv_kind = conv_kind
    self.A = A
    self.B = B
    self.C = C
    self.element_epilogue = element_epilogue
    self.epilogue_functor = epilogue_functor
    self.iterator_algorithm = iterator_algorithm
    self.stride_support = stride_support
    self.swizzling_functor = swizzling_functor

  #
  def core_name(self):
    ''' The basic operation kind is prefixed with a letter indicating the accumulation type. '''

    intermediate_type = ''

    if self.tile_description.math_instruction.opcode_class == OpcodeClass.TensorOp:
      inst_shape = "%d%d%d" % tuple(self.tile_description.math_instruction.instruction_shape)
      if self.tile_description.math_instruction.element_a != self.A.element and \
        self.tile_description.math_instruction.element_a != self.tile_description.math_instruction.element_accumulator:
        intermediate_type = DataTypeNames[self.tile_description.math_instruction.element_a]
    else:
      inst_shape = ''

    return "%s%s%s%s3d_%s" % (ShortDataTypeNames[self.tile_description.math_instruction.element_accumulator], \
      inst_shape, intermediate_type, ConvKindNames[self.conv_kind], IteratorAlgorithmNames[self.iterator_algorithm])

  #
  def extended_name(self):
    ''' Append data types if they differ from compute type. '''
    if self.C.element != self.tile_description.math_instruction.element_accumulator and \
      self.A.element != self.tile_description.math_instruction.element_accumulator:
      extended_name = "${element_c}_${core_name}_${element_a}"
    elif self.C.element == self.tile_description.math_instruction.element_accumulator and  \
      self.A.element != self.tile_description.math_instruction.element_accumulator:
      extended_name = "${core_name}_${element_a}"
    else:
      extended_name = "${core_name}"

    extended_name = SubstituteTemplate(extended_name, {
      'element_a': DataTypeNames[self.A.element],
      'element_c': DataTypeNames[self.C.element],
      'core_name': self.core_name()
      })

    return extended_name

  #
  def configuration_name(self):
    ''' The full procedural name indicates architecture, extended name, tile size, and layout. '''

    opcode_class_name = OpcodeClassNames[self.tile_description.math_instruction.opcode_class]
    
    threadblock = "%dx%d_%dx%d" % (
      self.tile_description.threadblock_shape[0],
      self.tile_description.threadblock_shape[1],
      self.tile_description.threadblock_shape[2],
      self.tile_description.stages
    )

    if self.stride_support == StrideSupport.Unity:
      configuration_name = "cutlass_${opcode_class}_${extended_name}_${threadblock}_unity_stride"
    else:
      configuration_name = "cutlass_${opcode_class}_${extended_name}_${threadblock}"

    return SubstituteTemplate(
      configuration_name,
      {
        'opcode_class': opcode_class_name,
        'extended_name': self.extended_name(),
        'threadblock': threadblock,
      }
    )

  #
  def procedural_name(self):
    ''' The full procedural name indicates architecture, extended name, tile size, and layout. '''
    return self.configuration_name()

###################################################################################################
#
# Emits single instances of a CUTLASS device-wide operator
#
###################################################################################################

class EmitConv3dInstance:
  def __init__(self):
    self.template = """
  // Conv3d${conv_kind_name} ${iterator_algorithm_name} kernel instance "${operation_name}"
  using ${operation_name}_base = 
  typename cutlass::conv::kernel::DefaultConv3d${conv_kind_name}<
    ${element_a}, 
    cutlass::layout::TensorNDHWC,
    ${element_b}, 
    cutlass::layout::TensorNDHWC,
    ${element_c}, 
    cutlass::layout::TensorNDHWC,
    ${element_accumulator},
    ${opcode_class},
    ${arch},
    cutlass::gemm::GemmShape<${threadblock_shape_m}, ${threadblock_shape_n}, ${threadblock_shape_k}>,
    cutlass::gemm::GemmShape<${warp_shape_m}, ${warp_shape_n}, ${warp_shape_k} >,
    cutlass::gemm::GemmShape<${instruction_shape_m}, ${instruction_shape_n}, ${instruction_shape_k}>,
    ${epilogue_functor}<
      ${element_c},
      ${epilogue_vector_length},
      ${element_accumulator},
      ${element_epilogue}
    >,
    ${swizzling_functor}, // cutlass::gemm::threadblock::GemmSplitKIdentityThreadblockSwizzle<>,
    ${stages},
    cutlass::arch::OpMultiplyAdd,
    ${iterator_algorithm},
    ${stride_support}
  >::Kernel;
"""


  def emit(self, operation):

    warp_shape = [int(operation.tile_description.threadblock_shape[idx] / operation.tile_description.warp_count[idx]) for idx in range(3)]

    epilogue_vector_length = int(min(operation.C.alignment * DataTypeSize[operation.C.element], 128) / DataTypeSize[operation.C.element])

    values = {
      'operation_name': operation.procedural_name(),
      'conv_kind': ConvKindTag[operation.conv_kind],
      'conv_kind_name': ConvKindNames[operation.conv_kind].capitalize(),
      'element_a': DataTypeTag[operation.A.element],
      'layout_a': LayoutTag[operation.A.layout],
      'element_b': DataTypeTag[operation.B.element],
      'layout_b': LayoutTag[operation.B.layout],
      'element_c': DataTypeTag[operation.C.element],
      'layout_c': LayoutTag[operation.C.layout],
      'element_accumulator': DataTypeTag[operation.tile_description.math_instruction.element_accumulator],
      'opcode_class': OpcodeClassTag[operation.tile_description.math_instruction.opcode_class],
      'arch': "cutlass::arch::Sm%d" % operation.arch,
      'threadblock_shape_m': str(operation.tile_description.threadblock_shape[0]),
      'threadblock_shape_n': str(operation.tile_description.threadblock_shape[1]),
      'threadblock_shape_k': str(operation.tile_description.threadblock_shape[2]),
      'warp_shape_m': str(warp_shape[0]),
      'warp_shape_n': str(warp_shape[1]),
      'warp_shape_k': str(warp_shape[2]),
      'instruction_shape_m': str(operation.tile_description.math_instruction.instruction_shape[0]),
      'instruction_shape_n': str(operation.tile_description.math_instruction.instruction_shape[1]),
      'instruction_shape_k': str(operation.tile_description.math_instruction.instruction_shape[2]),
      'epilogue_vector_length': str(epilogue_vector_length),
      'epilogue_functor': EpilogueFunctorTag[operation.epilogue_functor],
      'element_epilogue': str(DataTypeTag[operation.element_epilogue]),
      'swizzling_functor': SwizzlingFunctorTag[operation.swizzling_functor],
      'stages': str(operation.tile_description.stages),
      'iterator_algorithm': IteratorAlgorithmTag[operation.iterator_algorithm],
      'iterator_algorithm_name': IteratorAlgorithmNames[operation.iterator_algorithm].capitalize(),
      'stride_support': StrideSupportTag[operation.stride_support]
    }

    return SubstituteTemplate(self.template, values)

###################################################################################################
#
# Generator functions for all layouts
#
###################################################################################################

#
def GenerateConv3dTensorOp(manifest, tile_descriptions, min_cc, align = 128):

  for tile in tile_descriptions:
    for conv_kind in [ConvKind.Fprop, ConvKind.Dgrad, ConvKind.Wgrad]:

      if conv_kind == ConvKind.Fprop or (tile.math_instruction.element_accumulator in [DataType.f16, DataType.f32]):

        #
        output_types = [tile.math_instruction.element_a, tile.math_instruction.element_accumulator] \
          if DataTypeSize[tile.math_instruction.element_accumulator] == 32 \
          else [tile.math_instruction.element_accumulator,]

        for output_type in output_types:
          A = TensorDescription(tile.math_instruction.element_a, LayoutType.TensorNDHWC, int(align / DataTypeSize[tile.math_instruction.element_a]))
          B = TensorDescription(tile.math_instruction.element_b, LayoutType.TensorNDHWC, int(align / DataTypeSize[tile.math_instruction.element_b]))
          C = TensorDescription(output_type,  LayoutType.TensorNDHWC, max(1, int(align / DataTypeSize[output_type])))

          manifest.append(Conv3dOperation(conv_kind, min_cc, tile, A, B, C, tile.math_instruction.element_accumulator))

###################################################################################################
#
# Emitters functions for all targets
#
###################################################################################################

class EmitConv3dConfigurationLibrary:
  def __init__(self, operation_path, configuration_name):
    self.configuration_name = configuration_name
    self.configuration_path = os.path.join(operation_path, "%s.cu" % configuration_name)

    self.instance_emitter = EmitConv3dInstance()

    self.instance_template = """
${operation_instance}

// Derived class
struct ${operation_name} : 
  public ${operation_name}_base { };

///////////////////////////////////////////////////////////////////////////////////////////////////

"""
    self.header_template = """
/*
  Generated by conv3d_operation.py - Do not edit.
*/

///////////////////////////////////////////////////////////////////////////////////////////////////

#include "cutlass/cutlass.h"
#include "cutlass/library/library.h"
#include "cutlass/library/manifest.h"

#include "library_internal.h"
#include "conv3d_operation.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
"""

    self.configuration_header = """

namespace cutlass {
namespace library {

// Initialize all instances
void initialize_${configuration_name}(Manifest &manifest) {

"""

    self.configuration_instance = """
  using Operation_${operation_name} = cutlass::conv::device::ImplicitGemmConvolution<
    ${operation_name}>;

  manifest.append(new cutlass::library::Conv3dOperation<
    Operation_${operation_name}>(
      "${operation_name}"));

"""

    self.configuration_epilogue = """
}
"""
    self.epilogue_template = """

///////////////////////////////////////////////////////////////////////////////////////////////////

} // namespace library
} // namespace cutlass

///////////////////////////////////////////////////////////////////////////////////////////////////

"""

  #
  def __enter__(self):
    self.configuration_file = open(self.configuration_path, "w")
    self.configuration_file.write(SubstituteTemplate(self.header_template, {
      'configuration_name': self.configuration_name
      }))
    self.operations = []
    return self

  #
  def emit(self, operation):
    self.operations.append(operation)
    self.configuration_file.write(SubstituteTemplate(self.instance_template, {
      'configuration_name': self.configuration_name,
      'operation_name': operation.procedural_name(),
      'operation_instance': self.instance_emitter.emit(operation)
      }))

  #
  def __exit__(self, exception_type, exception_value, traceback):

    self.configuration_file.write(SubstituteTemplate(self.configuration_header, {
      'configuration_name': self.configuration_name
      }))

    for operation in self.operations:
      self.configuration_file.write(SubstituteTemplate(self.configuration_instance, {
        'configuration_name': self.configuration_name,
        'operation_name': operation.procedural_name()  
      }))

    self.configuration_file.write(self.configuration_epilogue)
    self.configuration_file.write(self.epilogue_template)
    self.configuration_file.close()


###################################################################################################
###################################################################################################




#
# \file generator.py
#
# \brief Generates the CUTLASS Library's instances
#

import enum
import os.path
import shutil

from library import *
from gemm_operation import *
from rank_k_operation import *
from rank_2k_operation import *
from trmm_operation import *
from symm_operation import *
from conv2d_operation import *  
from conv3d_operation import *  

###################################################################################################

class EmitOperationKindLibrary:
  def __init__(self, generated_path, kind, args):
    self.generated_path = generated_path
    self.kind = kind
    self.args = args
    self.emitters = {
      OperationKind.Gemm: EmitGemmConfigurationLibrary
      , OperationKind.Conv2d: EmitConv2dConfigurationLibrary  
      , OperationKind.Conv3d: EmitConv3dConfigurationLibrary  
      , OperationKind.RankK: EmitRankKConfigurationLibrary
      , OperationKind.Rank2K: EmitRank2KConfigurationLibrary
      , OperationKind.Trmm: EmitTrmmConfigurationLibrary
      , OperationKind.Symm: EmitSymmConfigurationLibrary
    }

    self.configurations = [];

    self.header_template ="""
/*
 Generated by manifest.py - Do not edit.
*/

#include "cutlass/cutlass.h"
#include "cutlass/library/library.h"
#include "cutlass/library/manifest.h"

namespace cutlass {
namespace library {

///////////////////////////////////////////////////////////////////////////////////////////////////

"""
    self.entry_template = """

//
// Entry point to construct operations
//
void initialize_all_${operation_name}_operations(Manifest &manifest) {
"""
    self.configuration_prototype_template = "void initialize_${configuration_name}(Manifest &manifest);\n"
    self.configuration_template ="  initialize_${configuration_name}(manifest);\n"

    self.epilogue_template ="""

}

///////////////////////////////////////////////////////////////////////////////////////////////////

} // namespace library
} // namespace cutlass

"""

  #
  def __enter__(self):
    self.operation_path = os.path.join(self.generated_path, OperationKindNames[self.kind])
    os.mkdir(self.operation_path)

    self.top_level_path = os.path.join(self.operation_path, "all_%s_operations.cu" % OperationKindNames[self.kind])

    self.top_level_file = open(self.top_level_path, "w")
    self.top_level_file.write(self.header_template)

    self.source_files = [self.top_level_path,]

    return self

  #
  def emit(self, configuration_name, operations):

    with self.emitters[self.kind](self.operation_path, configuration_name) as configuration_emitter:
      for operation in operations:
        configuration_emitter.emit(operation)
      
      self.source_files.append(configuration_emitter.configuration_path)

    self.configurations.append(configuration_name)
    self.top_level_file.write(SubstituteTemplate(self.configuration_prototype_template, {'configuration_name': configuration_name} ))

  #
  def __exit__(self, exception_type, exception_value, traceback):
    self.top_level_file.write(SubstituteTemplate(self.entry_template, {'operation_name': OperationKindNames[self.kind]}))

    for configuration_name in self.configurations:
      self.top_level_file.write(SubstituteTemplate(self.configuration_template, {'configuration_name': configuration_name}))

    self.top_level_file.write(self.epilogue_template)
    self.top_level_file.close()

class EmitInterfaceLibrary:
  def __init__(self, generated_path, operation_count, args):
    self.generated_path = generated_path
    self.args = args


    self.prototypes = []
    self.fn_calls = []
    self.operation_count = str(operation_count)

    self.top_level_hdr_template = '''
/*
 Generated by manifest.py - Do not edit.
*/
'''
    self.top_level_prologue = '''

#include "cutlass/library/library.h"
#include "cutlass/library/manifest.h"

namespace cutlass {
\tnamespace library {

${prototypes}

\t\tvoid initialize_all(Manifest &manifest) {
\t\t\tmanifest.reserve(${operation_count});\n\n
${fn_calls}
\t\t\t}

\t} // namespace library
} // namespace cutlass

'''

  #
  def __enter__(self):
    self.top_level_path = os.path.join(self.generated_path, 'initialize_all.cpp')

    self.top_level_file = open(self.top_level_path, "w")
    self.top_level_file.write(self.top_level_hdr_template)

    self.source_files = [self.top_level_path,]

    return self

  #
  def emit(self, operation_name):
    self.prototypes.append(SubstituteTemplate(
       "\t\tvoid initialize_all_${operation_kind}_operations(Manifest &manifest);",
       {'operation_kind': operation_name}))
    self.fn_calls.append(SubstituteTemplate(
       "\t\t\tinitialize_all_${operation_kind}_operations(manifest);",
       {'operation_kind': operation_name}))
    


  #
  def __exit__(self, exception_type, exception_value, traceback):
    self.top_level_file.write(SubstituteTemplate(self.top_level_prologue, {'prototypes':"\n".join(self.prototypes),
                                                                           'fn_calls':"\n".join(self.fn_calls),
                                                                           'operation_count': self.operation_count}))
    self.top_level_file.close()

###################################################################################################
###################################################################################################

class Options:
  def __init__(self):
    pass

###################################################################################################

#
class Manifest:

  #
  def __init__(self, args = None):
    self.operations = {}
    self.args = args
    self.operation_count = 0
    self.operations_by_name = {}
    
    self.kernel_filter = ''
    self.kernel_filter_list = []
    self.kernel_names = []
    self.operations_enabled = []
    self.selected_kernels = []
    self.ignore_kernel_names = []
    self.compute_capabilities = [50,]
    self.curr_build_dir = '.'
    self.filter_by_cc = True

    if self.args:
      self.kernel_filter = self.args.kernels
      self.curr_build_dir = args.curr_build_dir
      architectures = args.architectures.split(';') if len(args.architectures) else ['50',]
      self.compute_capabilities = [int(x) for x in architectures]
      
      if args.filter_by_cc in ['false', 'False', '0']:
        self.filter_by_cc = False

      if args.operations == 'all':
        self.operations_enabled = []
      else:
        operations_list = [
          OperationKind.Gemm
          , OperationKind.Conv2d    
          , OperationKind.Conv3d    
          , OperationKind.RankK
          , OperationKind.Trmm
          , OperationKind.Symm
        ] 
        self.operations_enabled = [x for x in operations_list if OperationKindNames[x] in args.operations.split(',')]

      if args.kernels == 'all':
        self.kernel_names = []
      else:
        self.kernel_names = [x for x in args.kernels.split(',') if x != '']

      self.ignore_kernel_names = [x for x in args.ignore_kernels.split(',') if x != '']

      if args.kernel_filter_file is None:
          self.kernel_filter_list = []
      else:
          self.kernel_filter_list = self.get_kernel_filters(args.kernel_filter_file)

  #
  def get_kernel_filters (self, kernelListFile):
    if os.path.isfile(kernelListFile):
        with open(kernelListFile, 'r') as fileReader:
            lines = [line.rstrip() for line in fileReader if not line.startswith("#")]
        
        lines = [re.compile(line) for line in lines if line]
        return lines
    else:
        return []

  #
  def filter_out_kernels(self, kernel_name, kernel_filter_list):

    for kernel_filter_re in kernel_filter_list:
        if kernel_filter_re.search(kernel_name) is not None:
            return True
        
    return False

    
  #
  def _filter_string_matches(self, filter_string, haystack):
    ''' Returns true if all substrings appear in the haystack in order'''
    substrings = filter_string.split('*')
    for sub in substrings:
      idx = haystack.find(sub)
      if idx < 0:
        return False
      haystack = haystack[idx + len(sub):]
    return True

  #
  def filter(self, operation):
    ''' Filtering operations based on various criteria'''

    # filter based on compute capability
    enabled = not (self.filter_by_cc)

    for cc in self.compute_capabilities:
      if cc >= operation.tile_description.minimum_compute_capability and \
         cc <= operation.tile_description.maximum_compute_capability and \
         (cc not in SharedMemPerCC or SharedMemPerCC[cc] >= CalculateSmemUsage(operation)):

        enabled = True
        break

    if not enabled:
      return False

    if len(self.operations_enabled) and not operation.operation_kind in self.operations_enabled:
      return False

    # eliminate duplicates
    if operation.procedural_name() in self.operations_by_name.keys():
      return False

    # Filter based on list of valid substrings
    if len(self.kernel_names):
      name = operation.procedural_name()
      enabled = False

      # compare against the include list
      for name_substr in self.kernel_names:
        if self._filter_string_matches(name_substr, name):
          enabled = True
          break

      # compare against the exclude list
      for name_substr in self.ignore_kernel_names:
        if self._filter_string_matches(name_substr, name):
          enabled = False
          break
          
    if len(self.kernel_filter_list) > 0:
        enabled = False
        if self.filter_out_kernels(operation.procedural_name(), self.kernel_filter_list):
            enabled = True

    # todo: filter based on compute data type
    return enabled
  #

  #
  def append(self, operation):
    ''' 
      Inserts the operation.

      operation_kind -> configuration_name -> []
    '''

    if self.filter(operation):
    
      self.selected_kernels.append(operation.procedural_name())

      self.operations_by_name[operation.procedural_name()] = operation

      # add the configuration
      configuration_name = operation.configuration_name()

      if operation.operation_kind not in self.operations.keys():
        self.operations[operation.operation_kind] = {}

      if configuration_name not in self.operations[operation.operation_kind].keys():
        self.operations[operation.operation_kind][configuration_name] = []

      self.operations[operation.operation_kind][configuration_name].append(operation)
      self.operation_count += 1
  #

  #
  def emit(self, target = GeneratorTarget.Library):

    operation_emitters = {
      GeneratorTarget.Library: EmitOperationKindLibrary 
    }
    interface_emitters = {
      GeneratorTarget.Library: EmitInterfaceLibrary 
    }

    generated_path = os.path.join(self.curr_build_dir, 'generated')

    # create generated/
    if os.path.exists(generated_path):
      shutil.rmtree(generated_path)

    os.mkdir(generated_path)

    source_files = []

    with interface_emitters[target](generated_path, self.operation_count, self.args) as iface_emitter:
      for operation_kind, configurations in self.operations.items():
        iface_emitter.emit(OperationKindNames[operation_kind])

      source_files += iface_emitter.source_files


    # for each operation kind, emit initializer for all configurations
    for operation_kind, configurations in self.operations.items():
      with operation_emitters[target](generated_path, operation_kind, self.args) as operation_kind_emitter:
        for configuration_name, operations in configurations.items():
          operation_kind_emitter.emit(configuration_name, operations)

        source_files += operation_kind_emitter.source_files

    # write the manifest.cmake file containing paths from all targets
    manifest_path = os.path.join(generated_path, "manifest.cmake")
    with open(manifest_path, "w") as manifest_file:

      target_name = 'cutlass_library_objs'

      target_text = SubstituteTemplate("""cutlass_target_sources(
  ${target_name}
  BATCH_SOURCES ON
  PRIVATE
""", { 'target_name': target_name})

      manifest_file.write(target_text)

      for source_file in source_files:
        manifest_file.write("    %s\n" % str(source_file.replace('\\', '/')))
      manifest_file.write(")")
  #

###################################################################################################



#################################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#################################################################################################
from pycutlass import *
import cutlass
from cuda import cuda
from cuda import nvrtc
import tempfile
import os
import ctypes

#
import json
import sqlite3


IncludeTemplate = r'''#include "${include}"
'''

#


class CompilationOptions:
    '''
    Compilation options.
    '''

    #
    def __init__(self, flags, architectures=[80], include_paths=[]):
        self.includes = []
        self.include_paths = include_paths
        self.flags = flags
        self.architectures = architectures

    def get_str(self):
        options = ""

        for flag in self.flags:
            options += " " + flag

        for incl in self.include_paths:
            options += ' --include-path=%s' % incl

        arch_list = "-arch="
        for idx, arch in enumerate(self.architectures):
            if idx:
                arch_list += ","
            arch_list += "sm_%d" % arch

        options += " " + arch_list
        return options

    #
    def get(self):
        options = []

        for flag in self.flags:
            options.append(bytes(str.encode(flag)))

        for incl in self.include_paths:
            options.append(bytes(str.encode('--include-path=%s' % incl)))

        arch_list = "-arch="
        for idx, arch in enumerate(self.architectures):
            if idx:
                arch_list += ","
            arch_list += "sm_%d" % arch

        options.append(bytes(str.encode(arch_list)))

        return options


def convertToBinaryData(filename):
    with open(filename, 'rb') as file:
        blobData = file.read()
    return blobData


def CDLLBin(host_binary):
    tempfile.tempdir = "./"
    temp_so = tempfile.NamedTemporaryFile(
        prefix='host_func', suffix='.so', delete=True)
    with open(temp_so.name, 'wb') as file:
        file.write(host_binary)
    host_lib = ctypes.CDLL(temp_so.name)
    return host_lib


class ArtifactManager:
    """
    Artifact manager
    """

    def __init__(self) -> None:
        try:
            connection = sqlite3.connect("./compiled_cache.db")
            cursor = connection.cursor()
            sqlite_create_table_query = """CREATE TABLE compiled_operations(op_key TEXT NOT NULL UNIQUE, cubin BLOB NOT NULL, hostbin BLOB NOT NULL, op_name TEXT NOT NULL, op_attrs TEXT NOT NULL)"""
            cursor.execute(sqlite_create_table_query)
            connection.commit()
            cursor.close()
        except:
            pass

        self.nvcc()
        self.compiled_cache_device = cutlass.CompileCache()
        self.compiled_cache_host = cutlass.CompileCache()
    
    def nvrtc(self):
        self.backend = "nvrtc"
        self.default_compile_options = [
            '-std=c++11', '-default-device',
        ]
    def nvcc(self):
        self.backend = "nvcc"
        self.default_compile_options = [
            '-std=c++11',
        ]
    def insert_operation(self, op_key, cubin, hostfile, op_name, op_attrs):
        connection = sqlite3.connect("./compiled_cache.db")
        cursor = connection.cursor()
        sqlite_insert_blob_query = """ INSERT OR IGNORE INTO compiled_operations (op_key, cubin, hostbin, op_name, op_attrs) VALUES (?, ?, ?, ?, ?)"""

        hostbin = convertToBinaryData(hostfile)

        data_tuple = (op_key, cubin, hostbin, op_name, json.dumps(op_attrs))

        cursor.execute(sqlite_insert_blob_query, data_tuple)
        connection.commit()
        cursor.close()

    def load_operation(self, op_key):
        connection = sqlite3.connect("./compiled_cache.db")
        cursor = connection.cursor()
        sqlite_fetch_blob_query = """SELECT * from compiled_operations where op_key = ?"""
        # try:
        cursor.execute(sqlite_fetch_blob_query, (op_key, ))
        record = cursor.fetchall()
        if len(record) == 0:
            return False
        for row in record:
            key, cubin_image, host_binary, operation_name, op_attr = row
            op_attr = json.loads(op_attr)
            err, module = cuda.cuModuleLoadData(cubin_image)
            if err != cuda.CUresult.CUDA_SUCCESS:
                raise RuntimeError('Cuda Error: {}'.format(err))

            err, kernel = cuda.cuModuleGetFunction(
                module, bytes(str.encode(operation_name)))
            self.compiled_cache_device.insert(key, kernel)

            compiled_host_fns = {}
            host_lib = CDLLBin(host_binary)

            func_name = operation_name + '_get_params'
            func = getattr(host_lib, func_name)
            func.restype = ctypes.POINTER(ctypes.c_char * op_attr[0])
            compiled_host_fns['get_args'] = func

            func_name = operation_name + '_shared_memory_size'
            func = getattr(host_lib, func_name)
            compiled_host_fns['shared_memory_capacity'] = func()

            for attr in op_attr:
                if isinstance(attr, str):
                    func_name = operation_name + '_' + attr
                    func = getattr(host_lib, func_name)
                    compiled_host_fns[attr] = func

            self.compiled_cache_host.insert(key, compiled_host_fns)
        return True

    def emit_compile_(self, operation_list, compilation_options):
        """
        Compile a list of kernels and store them into database
        """
        source_buffer_device = ""
        source_buffer_host = ""
        # 1. include
        includes = []
        for operation in operation_list:
            for incl in operation.emitter.includes:
                if incl not in includes:
                    includes.append(incl)

        includes_host = [
            "builtin_types.h", "device_launch_parameters.h", "stddef.h"] + includes
        for incl in includes:
            source_buffer_device += SubstituteTemplate(
                IncludeTemplate, {'include': incl})

        for incl in includes_host:
            if "/device/" not in incl:
                source_buffer_host += SubstituteTemplate(
                    IncludeTemplate, {'include': incl})

        # 2. Operations
        for operation in operation_list:
            source_buffer_device += operation.emit()
            source_buffer_host += operation.emit()
            values = {
                'operation_name': operation.name(),
                'operation_suffix': operation.emitter.operation_suffix
            }
            source_buffer_device += SubstituteTemplate(
                operation.KernelTemplate, values)
            source_buffer_host += SubstituteTemplate(
                operation.HostTemplate, values)

        if self.backend == "nvrtc":
            # 3. compile
            err, program = nvrtc.nvrtcCreateProgram(
                str.encode(source_buffer_device),
                bytes(str.encode("module.cu")),
                0, [], [])

            if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
                raise RuntimeError('NVRTC Error: {}'.format(err))

            # Compile program
            options = compilation_options.get()

            err, = nvrtc.nvrtcCompileProgram(program, len(options), options)
            if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:

                error_string = 'NVRTC Error: {}\n'.format(err)

                # Get log from compilation
                err, logSize = nvrtc.nvrtcGetProgramLogSize(program)
                if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
                    raise RuntimeError('NVRTC Error: {}'.format(err))

                log = b' ' * logSize
                err, = nvrtc.nvrtcGetProgramLog(program, log)
                if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
                    raise RuntimeError('NVRTC Error: {}'.format(err))

                raise RuntimeError(
                    error_string + log.decode() + source_buffer_device)

            # Get data from compilation
            err, dataSize = nvrtc.nvrtcGetCUBINSize(program)
            if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
                raise RuntimeError('NVRTC Error: {}'.format(err))

            cubin_image = b' ' * dataSize
            err, = nvrtc.nvrtcGetCUBIN(program, cubin_image)
            if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
                raise RuntimeError('NVRTC Error: {}'.format(err))
        else:  # with nvcc backend
            # emit code
            tempfile.tempdir = "./"
            temp_cu = tempfile.NamedTemporaryFile(
                prefix='kernel', suffix='.cu', delete=True)
            temp_cubin = tempfile.NamedTemporaryFile(
                prefix='kernel', suffix='.cubin', delete=True)
            with open(temp_cu.name, 'w') as file:
                file.write(source_buffer_device)

            # compile with nvcc
            cuda_install_path = os.getenv('CUDA_INSTALL_PATH')
            assert cuda_install_path is not None, "Environment variable 'CUDA_INSTALL_PATH' is not defined."
            cmd_template = "${cuda_install_path}/bin/nvcc ${options} -cubin ${srcfile} -o ${tarfile}"
            values = {
                "cuda_install_path": cuda_install_path,
                "options": compilation_options.get_str(),
                "srcfile": temp_cu.name,
                "tarfile": temp_cubin.name
            }
            cmd = SubstituteTemplate(cmd_template, values)
            os.system(cmd)

            # load the cubin image
            with open(temp_cubin.name, 'rb') as file:
                cubin_image = file.read()

        # compile the host code
        options = compilation_options.get()
        cmd = "echo '%s'|g++ -x c++ -fpermissive -w -fPIC" % source_buffer_host
        for opt in options:
            opt = opt.decode("utf-8")
            if opt not in ['-default-device', '-std=c++11', '-arch=sm_80', '-Xcicc', '-Xllc']:
                if '--include-path=' in opt:
                    cmd += " " + opt.replace('--include-path=', '-I')
                else:
                    cmd += " " + opt

        tempfile.tempdir = "./"
        temp = tempfile.NamedTemporaryFile(
            prefix='host_func', suffix='.so', delete=True)

        cmd += ' - -shared -o %s' % temp.name
        os.system(cmd)
        host_lib = ctypes.CDLL(temp.name)

        return cubin_image, host_lib, temp

    def add_module(self, operations, compile_options=None):
        """
        Insert a new compiled device module
        """
        if compile_options is None:
            cutlass_path = os.getenv('CUTLASS_PATH')
            assert cutlass_path is not None, "Environment variable 'CUTLASS_PATH' is not defined."
            cuda_install_path = os.getenv('CUDA_INSTALL_PATH')
            assert cuda_install_path is not None, "Environment variable 'CUDA_INSTALL_PATH' is not defined."
            architectures = []
            for operation in operations:
                if hasattr(operation, "tile_description"):
                    cc = operation.arch
                    if cc not in architectures:
                        architectures.append(cc)
            include_paths = [
                cuda_install_path + '/include',
                cutlass_path + '/include',
                cutlass_path + '/tools/util/include',
                cutlass_path + '/tools/library/scripts/pycutlass/src/cpp/include'
            ]
            compile_options = CompilationOptions(
                self.default_compile_options, architectures, include_paths)
        # save the cubin
        operation_key = []
        operation_list = []
        for operation in operations:
            # step 1: get kernel string as key
            key = operation.rt_module.emit() + operation.procedural_name() + self.backend
            # step 1: check if the operation is in cache
            compiled_kernel = self.compiled_cache_device.at(key)

            if compiled_kernel is None:
                hit = self.load_operation(key)
                if hit:
                    compiled_kernel = self.compiled_cache_device.at(key)
                    assert compiled_kernel is not None
            if compiled_kernel is not None:
                operation.rt_module.kernel = compiled_kernel
                compiled_host_fns = self.compiled_cache_host.at(key)
                assert compiled_host_fns is not None
                for key in compiled_host_fns.keys():
                    setattr(operation.rt_module, key, compiled_host_fns[key])
                operation.rt_module.initialize()
            else:
                operation_list.append(operation.rt_module)
                operation_key.append(key)
        if len(operation_list) > 0:
            cubin_image, host_lib, host_file = self.emit_compile_(
                operation_list, compile_options)

            err, module = cuda.cuModuleLoadData(cubin_image)
            if err != cuda.CUresult.CUDA_SUCCESS:
                raise RuntimeError('Cuda Error: {}'.format(err))

            operation_name = []
            operation_attr = []
            for operation, key in zip(operation_list, operation_key):
                # get device kernels
                err, operation.kernel = cuda.cuModuleGetFunction(
                    module,
                    bytes(str.encode(operation.name()))
                )
                operation_name.append(operation.name())
                self.compiled_cache_device.insert(key, operation.kernel)
                # get host functions
                compiled_host_fns = {}
                op_attr = []

                # get param size
                func_name = operation.name() + '_get_param_size'
                func = getattr(host_lib, func_name)
                param_size = func()

                func_name = operation.name() + '_get_params'
                func = getattr(host_lib, func_name)
                func.argtype = operation.argtype
                func.restype = ctypes.POINTER(ctypes.c_char * param_size)
                setattr(operation, 'get_args', func)
                compiled_host_fns['get_args'] = func

                # set shared memory size
                func_name = operation.name() + '_shared_memory_size'
                func = getattr(host_lib, func_name)
                setattr(operation, 'shared_memory_capacity', func())
                compiled_host_fns['shared_memory_capacity'] = func()
                # set the maximum dynamic shared size
                operation.initialize()

                # get extra functions
                op_attr.append(param_size)

                if hasattr(operation, "extra_funcs"):
                    for suffix in operation.extra_funcs:
                        func_name = operation.name() + '_' + suffix
                        func = getattr(host_lib, func_name)
                        setattr(operation, suffix, func)
                        compiled_host_fns[suffix] = func
                        op_attr.append(suffix)

                operation_attr.append(op_attr)
                self.compiled_cache_host.insert(key, compiled_host_fns)

            for key, operation_name, operation_attr in zip(operation_key, operation_name, operation_attr):
                self.insert_operation(
                    key, cubin_image, host_file.name, operation_name, operation_attr)



################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################
from pycutlass import *
from pycutlass.c_types import get_reduction_params
import cutlass
from cuda import cuda
try:
    import torch
    torch_available = True
except ImportError:
    torch_available = False
import numpy as np
from typing import Union
from cuda import cudart


class ReductionOperation:
    pass


class ReductionArguments:
    """
    Arguments of reduction
    """

    def __init__(self, operation: ReductionOperation,
                 problem_size: 'list[int]', partitions: int,
                 workspace: cuda.CUdeviceptr,
                 destination: 'Union[cuda.CUdeviceptr, np.ndarray, torch.Tensor]',
                 source: 'Union[cuda.CUdeviceptr, np.ndarray, torch.Tensor]', **kwargs) -> None:

        # tensor_C can be interpreted as the bias with bias=True in keyword args
        if "bias" in kwargs.keys():
            self.bias = kwargs["bias"]
        else:
            # by default, tensor_C is not bias
            self.bias = False

        self.operation = operation
        #: pointer to the workspace
        self.ptr_workspace = workspace

        #: number of split-k partitions
        self.partitions = partitions

        if isinstance(destination, np.ndarray):
            self.host_D = destination
            self.destination_buffer = NumpyFrontend.argument(destination, True)
            self.source_buffer = NumpyFrontend.argument(source, False)
            self.ptr_destination = cuda.CUdeviceptr(
                self.destination_buffer.ptr)
            self.ptr_source = cuda.CUdeviceptr(self.source_buffer.ptr)
        elif torch_available and isinstance(destination, torch.Tensor):
            self.ptr_destination = TorchFrontend.argument(destination)
            self.ptr_source = TorchFrontend.argument(source)
        elif isinstance(destination, cuda.CUdeviceptr):
            self.ptr_destination = destination
            self.ptr_source = source
        else:
            raise TypeError("unknown Type")

        self.problem_size = MatrixCoord_(
            problem_size[0], problem_size[1]
        )

        self.partition_stride = problem_size[0] * \
            problem_size[1] * DataTypeSize[operation.C.element] // 8

        if "output_op" in kwargs.keys():
            self.output_op = kwargs['output_op']
        else:
            self.output_op = self.operation.epilogue_type(1.0, 0.0)

        # get arguments
        self.get_arguments()

    @staticmethod
    def get_tensor_ref(extent: 'tuple[int]', device_ptr: cuda.CUdeviceptr, layout: cutlass.layout):
        if layout == cutlass.RowMajor:
            return TensorRef2D_(int(device_ptr), extent[1])
        else:
            raise ValueError("unknonwn layout type")

    def get_arguments(self):
        ref_workspace = ReductionArguments.get_tensor_ref(
            extent=[self.problem_size.row, self.problem_size.column],
            device_ptr=self.ptr_workspace, layout=cutlass.RowMajor)
        if self.bias:
            ref_source = ReductionArguments.get_tensor_ref(
                extent=[0, 0],
                device_ptr=self.ptr_source, layout=cutlass.RowMajor)
        else:
            ref_source = ReductionArguments.get_tensor_ref(
                extent=[self.problem_size.row, self.problem_size.column],
                device_ptr=self.ptr_source, layout=cutlass.RowMajor)

        ref_destination = ReductionArguments.get_tensor_ref(
            extent=[self.problem_size.row, self.problem_size.column],
            device_ptr=self.ptr_destination, layout=cutlass.RowMajor)


        self.c_arguments = self.operation.argument_type(
            self.problem_size, self.partitions,
            self.partition_stride, ref_workspace,
            ref_destination, ref_source,
            self.output_op
        )

        params_ = self.operation.rt_module.get_args(
            ctypes.byref(self.c_arguments))
        self.host_workspace = bytearray(params_.contents)

    def sync(self):
        err, = cudart.cudaDeviceSynchronize()
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError("CUDA Error %s" % str(err))

        if hasattr(self, "host_D"):
            err, = cuda.cuMemcpyDtoH(
                self.host_D, self.ptr_destination, self.host_D.size * self.host_D.itemsize)
            if err != cuda.CUresult.CUDA_SUCCESS:
                raise RuntimeError("CUDA Error %s" % str(err))

    def free(self):
        if hasattr(self, "destination_buffer"):
            del self.destination_buffer
        if hasattr(self, "source_buffer"):
            del self.source_buffer


class ReductionRT(ExecutableOperation):
    """
    ReductionRT manages the CUTLASS runtime components for reduction
    """
    KernelTemplate = r'''
extern "C"
__global__ void
${operation_name}(${operation_name}${operation_suffix}::Params params) {

  // Dynamic shared memory base pointer
  extern __shared__ int SharedStorageBase[];

  // Declare pointer to dynamic shared memory.
  ${operation_name}${operation_suffix}::SharedStorage *shared_storage =
      reinterpret_cast<${operation_name}${operation_suffix}::SharedStorage *>(SharedStorageBase);

  ${operation_name}${operation_suffix} op;

  op(params, *shared_storage);
}
    '''
    HostTemplate = r'''
extern "C" {
  // Get the size of params in bytes
  int ${operation_name}_get_param_size(){
    return sizeof(${operation_name}${operation_suffix}::Params);
  }

  // Get the size of dynamic shared memory in bytes
  int ${operation_name}_shared_memory_size() {
    return int(sizeof(${operation_name}${operation_suffix}::SharedStorage));
  }

  // Get the params as byte array
  char* ${operation_name}_get_params(${operation_name}${operation_suffix}::Params* params){
    char *bytes = ((char*)(params));
    char *output = new char[sizeof(${operation_name}${operation_suffix}::Params)];
    for (unsigned int i = 0; i < sizeof(${operation_name}${operation_suffix}::Params); i ++)
        output[i] = bytes[i];

    return output;
  }
}
    '''

    def __init__(self, operation: ReductionOperation):
        super().__init__(operation)

        self.operation: ReductionOperation = operation
        self.emitter = EmitReductionInstance('_type')

        self.elements_per_access = self.operation.count
        self.argument_type, self.epilogue_type = get_reduction_params(operation.epilogue_functor)
        self.argtype = [ctypes.POINTER(self.argument_type)]

    def emit(self):
        return self.emitter.emit(self.operation)

    def plan(self, arguments: ReductionArguments):
        block_shape = [self.operation.shape.column(
        ) // self.elements_per_access, self.operation.shape.row(), 1]
        grid_shape = [
            (arguments.problem_size.row + self.operation.shape.row() -
             1) // self.operation.shape.row(),
            (arguments.problem_size.column + self.operation.shape.column() -
                1) // self.operation.shape.column(),
            1
        ]
        return LaunchConfiguration(grid_shape, block_shape, self.shared_memory_capacity)

    def initialize(self):
        err, = cuda.cuFuncSetAttribute(
            self.kernel,
            attrib=cuda.CUfunction_attribute.CU_FUNC_ATTRIBUTE_MAX_DYNAMIC_SHARED_SIZE_BYTES,
            value=self.shared_memory_capacity)
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError('Cuda Error: {}'.format(err))


class ReductionOperation:
    """
    CUTLASS Reduction Operation
    shape: shape of CTA
    outputop: output operator
    r
    """

    def __init__(self, shape: cutlass.MatrixCoord, C: TensorDescription,
                 element_accumulator, element_workspace=None,
                 element_compute=None, epilogue_functor=None,
                 count: int = 1, partitions_per_stage: int = 4) -> None:
        """ Constructor
        """

        self.shape = shape
        #: epilogue functor (default: LinearCombination)
        self.epilogue_functor = epilogue_functor
        #: datatype of accumulator
        self.element_accumulator = element_accumulator

        if element_workspace is None:
            #: datatype of workspace
            self.element_workspace = element_accumulator
        else:
            #: datatype of workspace
            self.element_workspace = element_workspace

        if element_compute is None:
            #: datatype of workspace
            self.element_compute = element_accumulator
        else:
            #: datatype of workspace
            self.element_compute = element_compute

        #: datatype of output
        self.element_output = C.element

        #: operand C
        self.C: TensorDescription = C

        #: reduce op processing size
        self.count: int = count

        #: number of partitions to reduce per stage
        self.partitions_per_stage: int = partitions_per_stage

        self.rt_module: ReductionRT = ReductionRT(self)
        self.argument_type = self.rt_module.argument_type
        self.epilogue_type = self.rt_module.epilogue_type

    #
    def extended_name(self):
        extend_name = "${element_workspace}_${element_accumulator}_${element_compute}_${element_output}"

        return SubstituteTemplate(extend_name,
                                  {
                                      'element_workspace': DataTypeNames[self.element_workspace],
                                      'element_accumulator': DataTypeNames[self.element_accumulator],
                                      'element_compute': DataTypeNames[self.element_compute],
                                      'element_output': DataTypeNames[self.element_output]
                                  })

    #
    def configuration_name(self):
        ''' The full procedural name indicates architecture, extended name, tile size'''

        configuration_name = "cutlass_reduce_split_k_${extended_name}_${threadblock}"

        threadblock = "%dx%d" % (
            self.shape.row(),
            self.shape.column()
        )

        return SubstituteTemplate(
            configuration_name,
            {
                'extended_name': self.extended_name(),
                'threadblock': threadblock
            }
        )

    #
    def procedural_name(self):
        ''' The full procedural name indicates architeture, extended name, tile size'''
        return self.configuration_name()

    def run(self, arguments: ReductionArguments) -> cuda.CUresult:
        """
        Configure and launch the cuda kernel with input arguments
        """
        # get launch configuration
        launch_config = self.rt_module.plan(arguments)

        # get the host and device workspace
        host_workspace = arguments.host_workspace
        device_workspace = None

        # launch the kernel
        err = self.rt_module.run(
            host_workspace, device_workspace, launch_config)

        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError('CUDA Error %s' % str(err))

        return err


class EmitReductionInstance:
    def __init__(self, operation_suffix='') -> None:
        self.operation_suffix = operation_suffix
        self.includes = [
            "cutlass/cutlass.h",
            "cutlass/numeric_types.h",
            "cutlass/arch/arch.h",
            "cutlass/arch/mma.h",
            "cutlass/layout/matrix.h",
            "cutlass/gemm/device/gemm.h",
            "cutlass/gemm/device/gemm_universal_adapter.h",
            "cutlass/gemm/kernel/default_gemm_universal.h",
            "cutlass/reduction/kernel/reduce_split_k.h",
            "cutlass/reduction/thread/reduction_operators.h"
        ]
        self.template = """
// Reduction kernel instance
using ${operation_name}_base = 
typename cutlass::reduction::kernel::ReduceSplitK<
  cutlass::MatrixShape<${shape_row}, ${shape_column}>,
  ${epilogue_functor},
  cutlass::reduction::thread::ReduceAdd<
    ${element_accumulator},
    ${element_output},
    ${count}>,
  ${partition_per_stage}>;

struct ${operation_name}${operation_suffix}:
  public ${operation_name}_base { };
      """

    def emit(self, operation: ReductionOperation):

        epilogue_vector_length = int(min(
            operation.C.alignment * DataTypeSize[operation.C.element], 128) / DataTypeSize[operation.C.element])

        values = {
            'operation_name': operation.configuration_name(),
            'operation_suffix': self.operation_suffix,
            'shape_row': str(operation.shape.row()),
            'shape_column': str(operation.shape.column()),
            'epilogue_functor': operation.epilogue_functor.emit(),
            'element_output': DataTypeTag[operation.element_output],
            'epilogue_vector_length': str(epilogue_vector_length),
            'element_accumulator': DataTypeTag[operation.element_accumulator],
            'element_compute': DataTypeTag[operation.element_compute],
            'element_workspace': DataTypeTag[operation.element_workspace],
            'count': str(operation.count),
            'partition_per_stage': str(operation.partitions_per_stage)
        }

        return SubstituteTemplate(self.template, values)



################################################################################
#
# Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
################################################################################

import numpy as np
from cuda import cuda
from pycutlass.memory_manager import *
from typing import TYPE_CHECKING
try:
    import torch
    torch_available = True
except ImportError:
    torch_available = False
    if TYPE_CHECKING:
        import torch

try:
    import cupy as cp
    cupy_available = True
except ImportError:
    cupy_available = False
    if TYPE_CHECKING:
        import cupy as cp


class NumpyFrontend:
    """
    Frontend node for numpy
    """

    @staticmethod
    def argument(np_tensor: 'np.ndarray', is_output: 'bool') -> cuda.CUdeviceptr:
        """Convert the input numpy tensor to CUDA device pointer

        :param np_tensor: input numpy nd array
        :param is_output: whether the tensor is output

        :return: CUDA device pointer
        """
        # copy the data to device
        if is_output:
            return device_mem_alloc(np_tensor.size * np_tensor.itemsize)
        else:
            return todevice(np_tensor)


class TorchFrontend:
    """
    Frontend node for torch
    """

    @staticmethod
    def argument(torch_tensor: 'torch.Tensor') -> cuda.CUdeviceptr:
        """Convert the input torch tensor to CUDA device pointer

        :param torch_tensor: input torch tensor
        :param is_output: whether the tensor is output

        :return: CUDA device pointer
        """

        # check the device of torch_tensor
        if not torch_tensor.is_cuda:
            torch_tensor = torch_tensor.to("cuda")

        return cuda.CUdeviceptr(torch_tensor.data_ptr())


class CupyFrontend:
    """
    Frontend node for cupy
    """

    @staticmethod
    def argument(cupy_ndarray: 'cp.ndarray'):
        return cuda.CUdeviceptr(int(cupy_ndarray.data.ptr))



# test/unit/conv/device/conv2d_fprop_fixed_channels_f16nhwc_f16nhwc_f16nhwc_tensor_op_f32_sm80.cu
import pycutlass
from pycutlass.test import *
import unittest

def conv2d_fixed_channel_problemsizes(channels):
    problem_sizes = [
        cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(1, 8, 8, channels),
            cutlass.Tensor4DCoord(16, 3, 3, channels),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(2, 2),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ),
        cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(1, 224, 224, channels),
            cutlass.Tensor4DCoord(32, 7, 7, channels),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ),
        cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(1, 224, 224, channels),
            cutlass.Tensor4DCoord(64, 7, 7, channels),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(2, 2),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ),
        cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(1, 224, 224, channels),
            cutlass.Tensor4DCoord(64, 5, 5, channels),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ),
        cutlass.conv.Conv2dProblemSize(
            cutlass.Tensor4DCoord(1, 224, 224, channels),
            cutlass.Tensor4DCoord(64, 5, 5, channels),
            cutlass.Tensor4DCoord(1, 1, 1, 1),
            cutlass.MatrixCoord(2, 2),
            cutlass.MatrixCoord(1, 1),
            cutlass.conv.Mode.cross_correlation,
            1, 1
        ),
    ]

    return problem_sizes

class Conv2dFpropFixedChannelsF16NHWCF16NHWCF16HNWCTensorOpF32SM80(unittest.TestCase):
    def test_SM80_Device_Conv2d_Fprop_Fixed_Channels_ImplicitGemm_f16nhwc_f16nhwc_f16nhwc_tensor_op_f32_channels_8(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        A = TensorDescription(
            element=math_inst.element_a, 
            layout=cutlass.TensorNHWC,
            alignment=8)
        B = TensorDescription(
            element=math_inst.element_b, 
            layout=cutlass.TensorNHWC, 
            alignment=8)
        C = TensorDescription(
            element=cutlass.float16,
            layout=cutlass.TensorNHWC, 
            alignment=8)

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 64], stages=3, 
            warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, cutlass.float32)

        operation = Conv2dOperation(
            conv_kind=cutlass.conv.Operator.fprop, iterator_algorithm=cutlass.conv.IteratorAlgorithm.fixed_channels,
            arch=80, tile_description=tile_description, A=A, B=B, C=C, 
            stride_support=StrideSupport.Strided,
            epilogue_functor=epilogue_functor,
            swizzling_functor=cutlass.IdentitySwizzle1
        )
        
        self.assertTrue(test_all_conv2d(operation, conv2d_fixed_channel_problemsizes(8)))
    
    def test_SM80_Device_Conv2d_Fprop_Fixed_Channels_ImplicitGemm_f16nhwc_f16nhwc_f16nhwc_tensor_op_f32_channels_4(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        A = TensorDescription(
            element=math_inst.element_a, 
            layout=cutlass.TensorNHWC,
            alignment=4)
        B = TensorDescription(
            element=math_inst.element_b, 
            layout=cutlass.TensorNHWC, 
            alignment=4)
        C = TensorDescription(
            element=cutlass.float16,
            layout=cutlass.TensorNHWC, 
            alignment=8)

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 64], stages=3, 
            warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, cutlass.float32)

        operation = Conv2dOperation(
            conv_kind=cutlass.conv.Operator.fprop, iterator_algorithm=cutlass.conv.IteratorAlgorithm.fixed_channels,
            arch=80, tile_description=tile_description, A=A, B=B, C=C, 
            stride_support=StrideSupport.Strided,
            epilogue_functor=epilogue_functor,
            swizzling_functor=cutlass.IdentitySwizzle1
        )
        
        self.assertTrue(test_all_conv2d(operation, conv2d_fixed_channel_problemsizes(4)))
    
    def test_SM80_Device_Conv2d_Fprop_Fixed_Channels_ImplicitGemm_f16nhwc_f16nhwc_f16nhwc_tensor_op_f32_channels_2(self):
        math_inst = MathInstruction(
            instruction_shape=[16, 8, 16],
            element_a=cutlass.float16, element_b=cutlass.float16,
            element_accumulator=cutlass.float32, opcode_class=cutlass.OpClass.TensorOp,
            math_operation=MathOperation.multiply_add
        )

        A = TensorDescription(
            element=math_inst.element_a, 
            layout=cutlass.TensorNHWC,
            alignment=2)
        B = TensorDescription(
            element=math_inst.element_b, 
            layout=cutlass.TensorNHWC, 
            alignment=2)
        C = TensorDescription(
            element=cutlass.float16,
            layout=cutlass.TensorNHWC, 
            alignment=8)

        tile_description = TileDescription(
            threadblock_shape=[128, 128, 64], stages=3, 
            warp_count=[2, 2, 1],
            math_instruction=math_inst
        )

        epilogue_functor = LinearCombination(
            C.element, C.alignment, 
            math_inst.element_accumulator, cutlass.float32)

        operation = Conv2dOperation(
            conv_kind=cutlass.conv.Operator.fprop, iterator_algorithm=cutlass.conv.IteratorAlgorithm.fixed_channels,
            arch=80, tile_description=tile_description, A=A, B=B, C=C, 
            stride_support=StrideSupport.Strided,
            epilogue_functor=epilogue_functor,
            swizzling_functor=cutlass.IdentitySwizzle1
        )
        
        self.assertTrue(test_all_conv2d(operation, conv2d_fixed_channel_problemsizes(2)))


if __name__ == '__main__':
    pycutlass.get_memory_pool(2**26, 2**26)
    unittest.main()



import re


def SubstituteTemplate(template, values):
    text = template
    changed = True
    while changed:
        changed = False
        for key, value in values.items():
            regex = "\\$\\{%s\\}" % key
            newtext = re.sub(regex, value, text)
            if newtext != text:
                changed = True
            text = newtext
    return text

from pycutlass.type_hint import *
from pycutlass.tensor_ref import *
from pycutlass.operation import *
from pycutlass.epilogue import *
from pycutlass.parser import *
from pycutlass.compiler import ArtifactManager
from pycutlass.memory_manager import *
from pycutlass.arguments import *
from pycutlass.library import *
from pycutlass.c_types import *
from pycutlass.gemm_operation import *
from pycutlass.conv2d_operation import *
from pycutlass.compiler import *
from pycutlass.utils import *
from pycutlass.frontend import *
from pycutlass.reduction_operation import *
from pycutlass.compiler import *

# module-wide variables

import sys
this = sys.modules[__name__]

# artifact manager
this.compiler = ArtifactManager()

def get_memory_pool(init_pool_size=0, max_pool_size=2**34):
    this.memory_pool = PoolMemoryManager(
        init_pool_size=init_pool_size,
        max_pool_size=max_pool_size
    )
    return this.memory_pool
